import sys
import os
import glob
import builtins
import importlib
import re
import pathlib
import subprocess
from multiprocessing.pool import ThreadPool
import tempfile
import shutil
import logging

logger = logging.getLogger(__name__)


WORKER_SCRIPT = "worker.py"
PATCH_DIR = "patch"
PATCH_EXT = ".diff"
IMPORT_MODULES = re.compile(r"^\s*import\s+(.*)$")
FROM_IMPORT_MODULES = re.compile(r"^\s*from\s+(\S+)\s+import\s+(.*)$")
PYI_TO_PY = re.compile(r"\.pyi$")
RDKIT_MODULE_NAME = "rdkit"
RDBASE_MODULE_NAME = "rdBase"
INIT_PY = "__init__.py"

def rdkit_has_rdbase(rdkit_dir):
    """Returns True if rdkit_dir contains the rdBase Python module.

    Args:
        rdkit_dir (str): directory path

    Returns:
        bool: True if rdkit_dir contains the rdBase Python module
    """
    return any(f.startswith(RDBASE_MODULE_NAME) for f in os.listdir(rdkit_dir))

def purge_rdkit_source_dir_from_sys_path():
    """Remove the rdkit source dir from sys.path if present."""
    if os.path.isdir(RDKIT_MODULE_NAME) and not rdkit_has_rdbase(RDKIT_MODULE_NAME):
        abs_cwd = os.path.abspath(os.getcwd())
        indices_to_pop = sorted([i for i, p in enumerate(sys.path) if os.path.abspath(p) == abs_cwd], reverse=True)
        for i in indices_to_pop:
            sys.path.pop(i)

def find_rdkit_site_packages_path():
    site_packages_path = None
    for path_entry in sys.path:
        if not path_entry:
            continue
        rdkit_path = os.path.join(path_entry, RDKIT_MODULE_NAME)
        if os.path.isdir(rdkit_path) and rdkit_has_rdbase(rdkit_path):
            site_packages_path = path_entry
            break
    if site_packages_path is None:
        raise ValueError("Failed to find rdkit in PYTHONPATH")
    return site_packages_path

def find_rdkit_include_path():
    rdkit_include_path = None
    try:
        rdkit_include_path = find_rdkit_site_packages_path()
    except ValueError:
        pass
    if rdkit_include_path is not None:
        for i in range(3):
            rdkit_include_path = os.path.dirname(rdkit_include_path)
        rdkit_include_path = os.path.join(rdkit_include_path, "include")
        if not os.path.isdir(os.path.join(rdkit_include_path, RDKIT_MODULE_NAME)):
            rdkit_include_path = None
    return rdkit_include_path

def apply_patch(tempdir, patch_file):
    """Apply a patch file to the corresponding stub file."""
    cmd = ["git", "--git-dir=", "apply", patch_file]
    proc = subprocess.run(cmd, cwd=tempdir, capture_output=True)
    if proc.returncode:
        msg = proc.stderr.decode("utf-8") or "(no error message)"
        cmd_as_str = " ".join(cmd)
        logger.critical(f"Failed to apply patch {patch_file};\n\"{cmd_as_str}\" failed with:\n{msg}")

def patch_stubs(tempdir, src_entry):
    """Patch the stubs in src_entry if a diff fail is available."""
    patch_dir = os.path.join(os.path.dirname(__file__), PATCH_DIR)
    if not os.path.isdir(patch_dir):
        return
    for pyi in glob.glob(os.path.join(src_entry, "**/*.pyi")):
        pyi_rel = os.path.relpath(pyi, tempdir)
        patch_file = os.path.join(patch_dir, pyi_rel + PATCH_EXT)
        if os.path.exists(patch_file):
            apply_patch(tempdir, patch_file)

def copy_stubs(src_entry, outer_dirs):
    """Copy src_entry to each directory in outer_dirs.
    If src_entry is a directory it will be recursively copied.
    The src_entry path is stripped off any leading directory
    including the "rdkit" directory.

    Args:
        src_entry (str): full path to a file or directory
        outer_dirs (list[str]): list of destination directory paths
    """
    base_src_entry = None
    for part in pathlib.Path(src_entry).parts:
        if base_src_entry is not None:
            base_src_entry.append(part)
        if part == RDKIT_MODULE_NAME:
            base_src_entry = []
    for outer_dir in outer_dirs:
        dst_entry = os.path.join(outer_dir, *base_src_entry)
        if os.path.isdir(src_entry):
            shutil.copytree(src_entry, dst_entry, dirs_exist_ok=True)
        elif os.path.isfile(src_entry):
            shutil.copyfile(src_entry, dst_entry)

def run_worker(module_name):
    """Worker function passed to multiprocessing.Pool.map.

    Args:
        module_name (str): name of the Python module we
        are working on

    Returns:
        tuple[str, str]: tuple with stdout and stderr.
        The stderr contents are only reported if return code is nonzero
    """
    out = ""
    err = ""
    args = [
        "--tempdir", run_worker.tempdir,
        "--module-name", module_name,
    ]
    if run_worker.keep_incorrect_staticmethods:
        args.append("--keep-incorrect-staticmethods")
    cmd = run_worker.cmd + args + ["--outer-dirs"] + run_worker.outer_dirs
    proc = subprocess.run(cmd, capture_output=True)
    if proc.returncode:
        msg = proc.stderr.decode("utf-8") or "(no error message)"
        cmd_as_str = " ".join(cmd)
        err = f"\"{cmd_as_str}\" failed with:\n{msg}"
    if proc.stdout:
        out = proc.stdout.decode("utf-8")
    return out, err

def parse_modules_to_set(modules):
    """Convert comma-separated list of Python module names to a set.

    Args:
        modules (str): comma-separated list of Python module names

    Returns:
        set: set of unique Python module names
    """
    return {m.strip() for m in modules.split(",")}

def concat_parent_child_module(parent_module, child_module):
    """Convert "from x import y" to "import x.y".
    If y is "*", "from x import *" becomes "import x".

    Args:
        parent_module (str): from module
        child_module (str): imported module

    Returns:
        str: dot-separated concatenated import
    """
    if child_module != "*":
        parent_module += "." + child_module
    return parent_module

def clear_stubs(outer_dir):
    """Remove all files and directories from "rdkit-stubs"
    except CMakeLists.txt.

    Args:
        outer_dir (_type_): _description_
    """
    for entry in os.listdir(outer_dir):
        if entry in (
            "CMakeLists.txt",
            "gen_rdkit_stubs.out",
            "gen_rdkit_stubs.err"
        ):
            continue
        entry = os.path.join(outer_dir, entry)
        if os.path.isdir(entry):
            shutil.rmtree(entry)
        else:
            os.remove(entry)

def generate_stubs_internal(modules, outer_dirs, args):
    concurrency = min(args.concurrency, len(modules))
    with tempfile.TemporaryDirectory() as tempdir:
        src_dir = os.path.join(tempdir, RDKIT_MODULE_NAME)
        run_worker.cmd = [sys.executable, os.path.join(os.path.dirname(__file__), WORKER_SCRIPT)]
        run_worker.tempdir = tempdir
        run_worker.keep_incorrect_staticmethods = args.keep_incorrect_staticmethods
        run_worker.outer_dirs = outer_dirs
        with ThreadPool(concurrency) as pool:
            res = pool.map(run_worker, modules)
        concat_out, concat_err = tuple(zip(*res))
        concat_out = "\n".join(out for out in concat_out if out)
        concat_err = "\n".join(err for err in concat_err if err)
        if concat_err:
            logger.critical(concat_err)
        if concat_out and args.verbose:
            logger.warning(concat_out)
        if os.path.isdir(src_dir):
            patch_stubs(tempdir, src_dir)
            for f in os.listdir(src_dir):
                src_entry = os.path.join(src_dir, f)
                if os.path.exists(src_entry):
                    copy_stubs(src_entry, outer_dirs)

def generate_stubs(site_packages_path, args):
    """Generate RDKit stubs.

    Args:
        site_packages_path (str): full path from where RDKit modules are
        imported.
        output_dirs (list, optional): List of directories where a rdkit-stubs.
        directory is created. Defaults to [os.getcwd()].
        concurrency (int, optional): Number of CPUs used to generate stubs.
        Defaults to auto.
        verbose (bool, optional): Whether output should be verbose.
        Defaults to False.
    """
    output_dirs = args.output_dirs or [os.getcwd()]
    args.concurrency = args.concurrency or 1
    args.verbose = args.verbose or False
    args.keep_incorrect_staticmethods = args.keep_incorrect_staticmethods or False
    modules = {str(p.parent.relative_to(site_packages_path)).replace(os.sep, ".")
               for p in sorted(site_packages_path.joinpath(RDKIT_MODULE_NAME).rglob(INIT_PY))}
    outer_dirs = []
    for output_dir in output_dirs:
        outer_dir = os.path.join(output_dir, f"{RDKIT_MODULE_NAME}-stubs")
        outer_dirs.append(outer_dir)
        if os.path.isdir(outer_dir):
            clear_stubs(outer_dir)
        elif os.path.isfile(outer_dir):
            os.remove(outer_dir)
        if not os.path.isdir(outer_dir):
            os.makedirs(outer_dir)
    generate_stubs_internal(modules, outer_dirs, args)
    pyi_path = pathlib.Path(outer_dirs[0])
    pyi_files = {PYI_TO_PY.sub(".py", str(p.relative_to(pyi_path))) for p in pyi_path.rglob("*.pyi")}
    modules = {str(p.relative_to(site_packages_path.joinpath(RDKIT_MODULE_NAME)))
               for p in (
        sorted(site_packages_path.joinpath(RDKIT_MODULE_NAME).rglob("*.pyd")) +
        sorted(site_packages_path.joinpath(RDKIT_MODULE_NAME).rglob("*.so")) +
        sorted(site_packages_path.joinpath(RDKIT_MODULE_NAME).rglob("*.py"))
    ) if p.name != INIT_PY}.difference(pyi_files)
    modules = {RDKIT_MODULE_NAME + "." + os.path.splitext(p.replace(os.sep, "."))[0] for p in modules}
    generate_stubs_internal(modules, outer_dirs, args)

class PythonParameter:
    """Class to store Python function signature parameters."""

    ARG1 = "arg1"
    ARG_DIGIT = re.compile(r"^arg\d+$")

    def __init__(self, arg_type, arg_name, arg_default=None):
        self.arg_type = arg_type
        self.arg_name = arg_name
        self.arg_default = arg_default

    def as_str(self):
        """Return this parameter as a .pyi signature string parameter.

        Returns:
            str: parameter: type as string, followed by optional
            default parameter
        """
        res = f"{self.arg_name}: {self.arg_type}"
        if self.arg_default:
            res += f" = {self.arg_default}"
        return res

    def is_arg1(self):
        """Return True if this parameter is an arg1 parameter.

        Returns:
            bool: True if this parameter is an arg1 parameter
        """
        return self.arg_name == self.ARG1

    def is_arg_digit(self):
        """Return True if this parameter is an arg# parameter.

        Returns:
            bool: True if this parameter is an arg# parameter
        """
        return self.ARG_DIGIT.match(self.arg_name) is not None

    def rename(self, new_name):
        """Rename this parameter to new_name.

        Args:
            new_name (str): new parameter name
        """
        self.arg_name = new_name


class ProcessDocLines:
    """A class to pre-process docstrings before feeding them to pybind11_stubgen."""

    PY_SIGNATURE_ARG_REGEX = re.compile(r"^\((\S+)\)([^\=]+)\=?(.*)?$")
    DEF_REGEX = re.compile(r"^([^(]+)(\s*\(\s*).*(\s*\)\s*->\s*)[^:]+:(\s*)$")
    PURGE_CPP_OBJECT_ANGLE_BRACKETS = re.compile(r"^(.*)<(\S*\.)?(\S+)\s*object\s*at\s*\S+\s*>(.*)$")
    PURGE_OPEN_SQUARE_BRACKET = re.compile(r"\[(?!\])")
    PURGE_CLOSE_SQUARE_BRACKET = re.compile(r"(?<!\[)\]")
    PROTECTIONS = {
        "[": "__OPEN_SQUARE_BRACKET_TAG__",
        "]": "__CLOSE_SQUARE_BRACKET_TAG__",
        "=": "__EQUALS_TAG__"
    }
    CPP_PYTHONIC_RETURN_TYPES = {
        "_listSt6vectorIiSaIiEE": "typing.Sequence[typing.Sequence[int]]",
        "_ROConformerSeq": "typing.Sequence[rdkit.Chem.Conformer]",
        "_ROQAtomSeq": "typing.Sequence[rdkit.Chem.QueryAtom]",
        "_vectd": "typing.Sequence[double]",
        "_vecti": "typing.Sequence[int]",
        "_vectj": "typing.Sequence[int]",
        "_vectN5RDKit13Abbreviations22AbbreviationDefinitionE": "rdkit.Chem.rdAbbreviations.AbbreviationDefinition]",
        "_vectN5RDKit9Chirality10StereoInfoE": "typing.Sequence[rdkit.Chem.StereoInfo]",
        "_vectNSt7__cxx1112basic_stringIcSt11char_traitsIcESaIcEEE": "typing.Sequence[str]",
        "_vectSt6vectorIiSaIiEE": "typing.Sequence[typing.Sequence[int]]",
        "_vectSt6vectorIjSaIjEE": "typing.Sequence[typing.Sequence[int]]",
    }
    STD_MEMBER_FUNC_PARAM_NAMES = {
        "__add__": ("other",),
        "__and__": ("other",),
        "append": ("item",),
        "__call__": (),
        "__contains__": ("item",),
        "__copy__": (),
        "data": (),
        "__delitem__": ("item",),
        "__enter__": (),
        "__eq__": ("other",),
        "__exit__": ("exc_type", "exc_value", "traceback"),
        "extend": ("other",),
        "__getinitargs__": (),
        "__getitem__": ("item",),
        "__getstate__": (),
        "__iadd__": ("other",),
        "__iand__": ("other",),
        "__idiv__": ("other",),
        "__imul__": ("other",),
        "__init__": (),
        "__invert__": (),
        "__ior__": ("other",),
        "__isub__": ("other",),
        "__iter__": (),
        "key": (),
        "__len__": (),
        "__mul__": ("other",),
        "__ne__": ("other",),
        "__next__": (),
        "next": (),
        "__or__": ("other",),
        "__setitem__": ("item", "value"),
        "__setstate__": ("data",),
        "__str__": (),
        "__sub__": ("other",),
        "__truediv__": ("other",),
        "__xor__": ("other",),
    }
    OVERLOADED_FUNCTION_TAG = "Overloaded function."

    def __init__(self, module_name, keep_incorrect_staticmethods=False):
        self.module_name = module_name
        self.keep_incorrect_staticmethods = keep_incorrect_staticmethods
        self.num_overloads = 0
        self.overload_num = 0
        self.top_signature = None

    @classmethod
    def protect_quoted_square_brackets_and_equals(cls, arg):
        """Replaces [, ], = with string tags.

        Args:
            arg (str): raw single Python signature parameter

        Returns:
            str: single Python signature parameter with symbols
            replaced by string tags
        """
        open_quote = False
        protected_arg = ""
        for c in arg:
            if c == "'":
                open_quote ^= True
            elif open_quote:
                replacement = cls.PROTECTIONS.get(c, None)
                if replacement is not None:
                    c = replacement
            protected_arg += c
        return protected_arg

    @classmethod
    def deprotect_quoted_square_brackets_and_equals(cls, arg):
        """Restores [, ], = which were replaced by string tags.

        Args:
            arg (str): single Python signature parameter with symbols
            replaced by string tags

        Returns:
            str: single Python signature parameter with original symbols
        """
        for k, v in cls.PROTECTIONS.items():
            arg = arg.replace(v, k)
        return arg

    @classmethod
    def process_py_signature_arg(cls, arg):
        """Process single Python signature parameter.

        Args:
            arg (str): raw single Python signature parameter

        Returns:
            str: processed single Python signature parameter
        """
        arg = cls.PURGE_CPP_OBJECT_ANGLE_BRACKETS.sub(r"\1\3()\4", arg)
        arg = cls.protect_quoted_square_brackets_and_equals(arg)
        arg = cls.PURGE_OPEN_SQUARE_BRACKET.sub("", arg)
        arg = cls.PURGE_CLOSE_SQUARE_BRACKET.sub("", arg)
        arg = arg.replace(" ", "")
        py_signature_arg_match = cls.PY_SIGNATURE_ARG_REGEX.match(arg)
        assert py_signature_arg_match
        py_signature_arg_type = py_signature_arg_match.group(1)
        py_signature_arg_name = py_signature_arg_match.group(2)
        py_signature_arg_default = cls.deprotect_quoted_square_brackets_and_equals(py_signature_arg_match.group(3))
        return PythonParameter(py_signature_arg_type, py_signature_arg_name, py_signature_arg_default)

    @classmethod
    def find_def_match(cls, src_line):
        """Find Python function def in src_line.

        Args:
            src_line (str): single docstring line

        Returns:
            re.Match|None: the re.Match or None if no match
        """
        return cls.DEF_REGEX.match(src_line)

    def convert_to_valid_type(self, py_signature_ret):
        """Takes as input a return type which may be either a builtin
        or an RDKit class type. Builtins are returned unchanged, while
        RDKit class types are prefixed with the appropriate module such
        that an import can be added at the top of the .pyi file by
        pybind11_stubgen and the return type is correctly recognized
        by Pyright.

        Args:
            py_signature_ret (str): raw return type

        Returns:
            str: return type prefixed with the relevant Python module
            if needed
        """
        res = self.CPP_PYTHONIC_RETURN_TYPES.get(py_signature_ret, py_signature_ret)
        is_valid_type = hasattr(builtins, py_signature_ret)
        module_name_tmp = self.module_name
        while module_name_tmp and not is_valid_type:
            module_tmp = importlib.import_module(module_name_tmp)
            is_valid_type = hasattr(module_tmp, res)
            if is_valid_type:
                res = f"{module_name_tmp}.{py_signature_ret}"
            else:
                module_name_tmp = ".".join(module_name_tmp.split(".")[:-1])
        return res

    def correct_function_args(self, func_name, args):
        """Correct Python signatures.

        In spite of our efforts, boost::python will still generate
        incorrect docstrings for certain class methods, i.e., with
        "arg1" instead of the "self" parameter. This happens, among
        others, for those generated with boost::python::make_constructor.
        We cannot patch the C++ sources adding the "self" parameter
        as a boost::python::arg or the compiler will issue an error.
        Therefore, the only option is to post-process the generated
        .pyi files and replace "arg1" with "self". When we do so
        and the parameters which follow are "arg#", we also renumber
        them accordingly.

        Args:
            func_name (str): name of the function whose signature
            may needs to be corrected
            args (list[PythonParameter]): list of PythonParameter
            instances that may need to be corrected

        Returns:
            list[PythonParameter]: corrected list of PythonParameter
            instances
        """
        param_names = self.STD_MEMBER_FUNC_PARAM_NAMES.get(func_name, None)
        if param_names is not None and args and args[0].is_arg1():
            param_names_with_self = ["self", *param_names]
            param_idx = 0
            for i, arg in enumerate(args):
                param_name = None
                if i <= len(param_names):
                    param_name = param_names_with_self[i]
                elif arg.is_arg_digit():
                    param_idx += 1
                    param_name = f"arg{param_idx}"
                if param_name is not None:
                    arg.rename(param_name)
        return args

    def process_src_line(self, src_line):
        """Process single docstring line.
        Function def lines are processed, others are returned with
        no change.
        Currently, processing involves:
        * Processing function parameters to assign a type
          (extracted from the docstring itself)
        * Processing return type to prepend the Pythonmodule
          where the type is defined as needed
        * For overloaded functiona and methods, the first overload becomes
          the top-line signature, followed by a line with the
          OVERLOADED_FUNCTION_TAG. Then all the overloads follow as
          listed in the docstring prepended by a "#." tag where
          # is the 1-based overload number.
          This is the overload format that pybind11_stubgen expects
          and that will trigger the addition of the @typing.overload
          decorators as well as the per-overload docstring trimming

        Args:
            src_line (str): single docstring line

        Returns:
            str: processed docstring line
        """
        def_match = self.find_def_match(src_line)
        if def_match:
            func_name = def_match.group(1)
            overload_prefix = ""
            self.overload_num += 1
            if self.num_overloads > 1:
                overload_prefix = f"{self.overload_num}. "
            func_open_bracket = def_match.group(2)
            func_end_bracket_and_arrow = def_match.group(3)
            func_colon_to_end = def_match.group(4)
            py_signature_regex = re.compile(r"^\s*" + func_name + r"\s*\((.*)\)\s*->\s*(\S+)\s*:\s*$")
            py_signature_match = py_signature_regex.match(src_line)
            if py_signature_match:
                py_signature_args = py_signature_match.group(1)
                if py_signature_args:
                    py_signature_args = py_signature_args.split(", ")
                else:
                    py_signature_args = []
                py_signature_ret = py_signature_match.group(2)
                py_signature_ret = self.convert_to_valid_type(py_signature_ret)
                processed_args = [self.process_py_signature_arg(py_signature_arg) for py_signature_arg in py_signature_args]
                if not self.keep_incorrect_staticmethods:
                    processed_args = self.correct_function_args(func_name, processed_args)
                processed_args = ", ".join(arg.as_str() for arg in processed_args)
                src_line = f"{func_name}{func_open_bracket}{processed_args}{func_end_bracket_and_arrow}{py_signature_ret}{func_colon_to_end}"
                if self.top_signature is None:
                    self.top_signature = src_line
                if overload_prefix:
                    src_line = overload_prefix + src_line
        return src_line

    def process_doc_lines(self, doc_lines):
        """Process the raw docstring lines.
        * Count the number of overloads for the function described
          in the docstring
        * Trim any empty lines at the beginning of the docstring,
          as pybind11_stubgen does expects no empty lines
          at the top of the docstring or it will misbehave
        * If the function has >1 overload, prepend the first overload
          as top signature followed by a 2nd line bearing the
          OVERLOADED_FUNCTION_TAG string

        Args:
            doc_lines (list[str]): raw docstring lines

        Returns:
            list[str]: processed docstring lines
        """
        self.num_overloads = len(tuple(filter(None, [self.find_def_match(doc_line) for doc_line in doc_lines])))
        doc_lines = list(map(self.process_src_line, doc_lines))
        i = 0
        for i, doc_line in enumerate(doc_lines):
            if doc_line:
                break
        for _ in range(i):
            doc_lines.pop(0)
        if self.num_overloads > 1:
            doc_lines.insert(0, self.top_signature)
            doc_lines.insert(1, self.OVERLOADED_FUNCTION_TAG)
        return doc_lines

    @classmethod
    def process(cls, module_name, doc_lines, keep_incorrect_staticmethods=False):
        """Process the raw docstring lines.
        This is a convenience static function that creates an instance
        of ProcessDocLines and calls process_doc_lines() on it.

        Args:
            module_name (str): fully qualified Python module
            name the docstring belongs to
            doc_lines (list[str]): raw docstring lines
            keep_incorrect_staticmethods (bool): if true, incorrectly
            typed staticmethods are left unmodified

        Returns:
            list[str]: processed docstring lines
        """
        instance = cls(module_name, keep_incorrect_staticmethods)
        return instance.process_doc_lines(doc_lines)



#
#  Copyright (c) 2013, Novartis Institutes for BioMedical Research Inc.
#  Copyright (c) 2021, Greg Landrum
#  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#     * Neither the name of Novartis Institutes for BioMedical Research Inc.
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Created by Sereina Riniker, Aug 2013

import copy
import math

try:
  import matplotlib
  from matplotlib import cm
  from matplotlib.colors import LinearSegmentedColormap
except ImportError:
  cm = None
except RuntimeError:
  cm = None

import numpy

from rdkit import Chem, DataStructs, Geometry
from rdkit.Chem import Draw, rdDepictor
from rdkit.Chem import rdFingerprintGenerator
from rdkit.Chem import rdMolDescriptors as rdMD
from rdkit.Chem.Draw import rdMolDraw2D


def _DeleteFpInfoAttr(mol):
  if hasattr(mol, '_fpInfo'):
    delattr(mol, '_fpInfo')


def GetAtomicWeightsForFingerprint(refMol, probeMol, fpFunction, metric=DataStructs.DiceSimilarity):
  """
    Calculates the atomic weights for the probe molecule
    based on a fingerprint function and a metric.

    Parameters:
      refMol -- the reference molecule
      probeMol -- the probe molecule
      fpFunction -- the fingerprint function
      metric -- the similarity metric

    Note:
      If fpFunction needs additional parameters, use a lambda construct
    """
  _DeleteFpInfoAttr(probeMol)
  _DeleteFpInfoAttr(refMol)

  refFP = fpFunction(refMol, -1)
  baseSimilarity = metric(refFP, fpFunction(probeMol, -1))

  # loop over atoms
  weights = [
    baseSimilarity - metric(refFP, fpFunction(probeMol, atomId))
    for atomId in range(probeMol.GetNumAtoms())
  ]

  _DeleteFpInfoAttr(probeMol)
  _DeleteFpInfoAttr(refMol)
  return weights


def GetAtomicWeightsForModel(probeMol, fpFunction, predictionFunction):
  """
    Calculates the atomic weights for the probe molecule based on
    a fingerprint function and the prediction function of a ML model.

    Parameters:
      probeMol -- the probe molecule
      fpFunction -- the fingerprint function
      predictionFunction -- the prediction function of the ML model
    """

  _DeleteFpInfoAttr(probeMol)
  baseProba = predictionFunction(fpFunction(probeMol, -1))

  # loop over atoms
  weights = [
    baseProba - predictionFunction(fpFunction(probeMol, atomId))
    for atomId in range(probeMol.GetNumAtoms())
  ]

  _DeleteFpInfoAttr(probeMol)
  return weights


def GetStandardizedWeights(weights):
  """
    Normalizes the weights,
    such that the absolute maximum weight equals 1.0.

    Parameters:
      weights -- the list with the atomic weights
    """

  maxAbsWeight = max(math.fabs(w) for w in weights)

  if maxAbsWeight > 0:
    return [w / maxAbsWeight for w in weights], maxAbsWeight
  return weights, maxAbsWeight


def GetSimilarityMapFromWeights(mol, weights, draw2d, colorMap=None, scale=-1, size=(250, 250),
                                sigma=None, coordScale=1.5, step=0.01, colors='k', contourLines=10,
                                alpha=0.5, gridResolution=0.1, extraGridPadding=0.5,
                                useFillThreshold=False, fillThreshold=0.01,
                                fillThresholdIsFraction=True, **kwargs):
  """
    Generates the similarity map for a molecule given the atomic weights.

    Parameters:
      mol -- the molecule of interest
      weights -- the weights to use
      draw2d -- the RDKit drawing object
      colorMap -- the matplotlib color map scheme, default is custom PiWG color map
      scale -- the scaling: scale < 0 -> the absolute maximum weight is used as maximum scale
                            scale = double -> this is the maximum scale
      size -- the size of the figure
      sigma -- the sigma for the Gaussians
      coordScale -- scaling factor for the coordinates
      step -- the step for calcAtomGaussian
      colors -- color of the contour lines
      contourLines -- if integer number N: N contour lines are drawn
                      if list(numbers): contour lines at these numbers are drawn
      alpha -- the alpha blending value for the contour lines
      gridResolution -- the resolution of the grid
      extraGridPadding -- the extra padding of the grid
      useFillThreshold -- use a magnitude threshold to determine if a grid box is filled
      fillThreshold -- the magnitude threshold for filling grid boxes
      fillThresholdIsFraction -- if True, the fillThreshold is a fraction of the data range
      kwargs -- additional arguments for drawing
    """
  if mol.GetNumAtoms() < 2:
    raise ValueError("too few atoms")

  if draw2d is None:
    raise ValueError("the draw2d argument must be provided")
  mol = rdMolDraw2D.PrepareMolForDrawing(mol, addChiralHs=False)
  if not mol.GetNumConformers():
    rdDepictor.Compute2DCoords(mol)
  if sigma is None:
    if mol.GetNumBonds() > 0:
      bond = mol.GetBondWithIdx(0)
      idx1 = bond.GetBeginAtomIdx()
      idx2 = bond.GetEndAtomIdx()
      sigma = 0.3 * (mol.GetConformer().GetAtomPosition(idx1) -
                     mol.GetConformer().GetAtomPosition(idx2)).Length()
    else:
      sigma = 0.3 * (mol.GetConformer().GetAtomPosition(0) -
                     mol.GetConformer().GetAtomPosition(1)).Length()
    sigma = round(sigma, 2)

  sigmas = [sigma] * mol.GetNumAtoms()
  locs = []
  for i in range(mol.GetNumAtoms()):
    p = mol.GetConformer().GetAtomPosition(i)
    locs.append(Geometry.Point2D(p.x, p.y))
  draw2d.ClearDrawing()
  ps = Draw.ContourParams()
  ps.fillGrid = True
  ps.gridResolution = gridResolution
  ps.extraGridPadding = extraGridPadding
  ps.useFillThreshold = useFillThreshold
  ps.fillThreshold = fillThreshold
  ps.fillThresholdIsFraction = fillThresholdIsFraction

  if colorMap is not None:
    if cm is not None and isinstance(colorMap, type(cm.Blues)):
      # it's a matplotlib colormap:
      clrs = [tuple(x) for x in colorMap([0, 0.5, 1])]
    elif type(colorMap) == str:
      if cm is None:
        raise ValueError("cannot provide named colormaps unless matplotlib is installed")
      clrs = [tuple(x) for x in matplotlib.colormaps[colorMap]([0, 0.5, 1])]
    else:
      clrs = [colorMap[0], colorMap[1], colorMap[2]]
    ps.setColourMap(clrs)

  Draw.ContourAndDrawGaussians(draw2d, locs, weights, sigmas, nContours=contourLines, params=ps)
  draw2d.drawOptions().clearBackground = False
  draw2d.DrawMolecule(mol)
  return draw2d


def GetSimilarityMapForFingerprint(refMol, probeMol, fpFunction, draw2d,
                                   metric=DataStructs.DiceSimilarity, **kwargs):
  """
    Generates the similarity map for a given reference and probe molecule,
    fingerprint function and similarity metric.

    Parameters:
      refMol -- the reference molecule
      probeMol -- the probe molecule
      fpFunction -- the fingerprint function
      metric -- the similarity metric.
      kwargs -- additional arguments for drawing
    """

  weights = GetAtomicWeightsForFingerprint(refMol, probeMol, fpFunction, metric)
  weights, maxWeight = GetStandardizedWeights(weights)
  draw2d = GetSimilarityMapFromWeights(probeMol, weights, draw2d, **kwargs)
  return draw2d, maxWeight

def GetAtomicWeightsForFingerprintGenerator(refMol, probeMol, fpg, useCounts=False, metric=DataStructs.DiceSimilarity):
  """
    Calculates the atomic weights for the probe molecule
    based on a fingerprint function and a metric.

    Parameters:
      refMol -- the reference molecule
      probeMol -- the probe molecule
      fpg -- the fingerprint generator
      metric -- the similarity metric

    """
  if not useCounts:  
    refFP = fpg.GetFingerprint(refMol)
  else:
    refFP = fpg.GetCountFingerprint(refMol)
  probeAo = rdFingerprintGenerator.AdditionalOutput()
  probeAo.AllocateAtomsPerBit()
  if not useCounts:
    probeFP = fpg.GetFingerprint(probeMol, additionalOutput=probeAo)
  else:
    probeFP = fpg.GetCountFingerprint(probeMol, additionalOutput=probeAo)

  baseSimilarity = metric(refFP, probeFP)

  dmat = Chem.GetDistanceMatrix(probeMol)
  
  atomsPerBit = probeAo.GetAtomsPerBit()
  atomFps = [copy.deepcopy(probeFP) for i in range(probeMol.GetNumAtoms())]
  for bit,entries in atomsPerBit.items():
    for entry in entries:
      for aid in entry:
        if not useCounts:
          atomFps[aid].UnSetBit(bit)
        else:
          atomFps[aid][bit] -= 1
  
  weights = []
  for atomId in range(probeMol.GetNumAtoms()):
    weights.append(baseSimilarity - metric(refFP, atomFps[atomId ]))
  return weights


def GetSimilarityMapForFingerprintGenerator(refMol, probeMol, fpg, draw2d,
                                   metric=DataStructs.DiceSimilarity, useCounts=False,
                                   **kwargs):
  """
    Generates the similarity map for a given reference and probe molecule,
    fingerprint function and similarity metric.

    Parameters:
      refMol -- the reference molecule
      probeMol -- the probe molecule
      fpg -- the fingerprint generator
      metric -- the similarity metric.
      kwargs -- additional arguments for drawing
    """

  weights = GetAtomicWeightsForFingerprintGenerator(refMol, probeMol, fpg, metric=metric, useCounts=useCounts)
  weights, maxWeight = GetStandardizedWeights(weights)
  draw2d = GetSimilarityMapFromWeights(probeMol, weights, draw2d, **kwargs)
  return draw2d, maxWeight


def GetSimilarityMapForModel(probeMol, fpFunction, predictionFunction, draw2d, **kwargs):
  """
    Generates the similarity map for a given ML model and probe molecule,
    and fingerprint function.

    Parameters:
      probeMol -- the probe molecule
      fpFunction -- the fingerprint function
      predictionFunction -- the prediction function of the ML model
      kwargs -- additional arguments for drawing
    """

  weights = GetAtomicWeightsForModel(probeMol, fpFunction, predictionFunction)
  weights, maxWeight = GetStandardizedWeights(weights)
  fig = GetSimilarityMapFromWeights(probeMol, weights, draw2d, **kwargs)
  return fig, maxWeight


apDict = {}
apDict['normal'] = lambda m, bits, minl, maxl, bpe, ia, **kwargs: rdMD.GetAtomPairFingerprint(
  m, minLength=minl, maxLength=maxl, ignoreAtoms=ia, **kwargs)
apDict['hashed'] = lambda m, bits, minl, maxl, bpe, ia, **kwargs: rdMD.GetHashedAtomPairFingerprint(
  m, nBits=bits, minLength=minl, maxLength=maxl, ignoreAtoms=ia, **kwargs)
apDict[
  'bv'] = lambda m, bits, minl, maxl, bpe, ia, **kwargs: rdMD.GetHashedAtomPairFingerprintAsBitVect(
    m, nBits=bits, minLength=minl, maxLength=maxl, nBitsPerEntry=bpe, ignoreAtoms=ia, **kwargs)


# usage:   lambda m,i: GetAPFingerprint(m, i, fpType, nBits, minLength, maxLength, nBitsPerEntry)
def GetAPFingerprint(mol, atomId=-1, fpType='normal', nBits=2048, minLength=1, maxLength=30,
                     nBitsPerEntry=4, **kwargs):
  """
    Calculates the atom pairs fingerprint with the torsions of atomId removed.

    Parameters:
      mol -- the molecule of interest
      atomId -- the atom to remove the pairs for (if -1, no pair is removed)
      fpType -- the type of AP fingerprint ('normal', 'hashed', 'bv')
      nBits -- the size of the bit vector (only for fpType='bv')
      minLength -- the minimum path length for an atom pair
      maxLength -- the maxmimum path length for an atom pair
      nBitsPerEntry -- the number of bits available for each pair
    """
  if fpType not in ['normal', 'hashed', 'bv']:
    raise ValueError("Unknown Atom pairs fingerprint type")
  if atomId < 0:
    return apDict[fpType](mol, nBits, minLength, maxLength, nBitsPerEntry, 0, **kwargs)
  if atomId >= mol.GetNumAtoms():
    raise ValueError("atom index greater than number of atoms")
  return apDict[fpType](mol, nBits, minLength, maxLength, nBitsPerEntry, [atomId], **kwargs)


ttDict = {}
ttDict['normal'] = lambda m, bits, ts, bpe, ia, **kwargs: rdMD.GetTopologicalTorsionFingerprint(
  m, targetSize=ts, ignoreAtoms=ia, **kwargs)
ttDict[
  'hashed'] = lambda m, bits, ts, bpe, ia, **kwargs: rdMD.GetHashedTopologicalTorsionFingerprint(
    m, nBits=bits, targetSize=ts, ignoreAtoms=ia, **kwargs)
ttDict[
  'bv'] = lambda m, bits, ts, bpe, ia, **kwargs: rdMD.GetHashedTopologicalTorsionFingerprintAsBitVect(
    m, nBits=bits, targetSize=ts, nBitsPerEntry=bpe, ignoreAtoms=ia, **kwargs)


# usage:   lambda m,i: GetTTFingerprint(m, i, fpType, nBits, targetSize)
def GetTTFingerprint(mol, atomId=-1, fpType='normal', nBits=2048, targetSize=4, nBitsPerEntry=4,
                     **kwargs):
  """
    Calculates the topological torsion fingerprint with the pairs of atomId removed.

    Parameters:
      mol -- the molecule of interest
      atomId -- the atom to remove the torsions for (if -1, no torsion is removed)
      fpType -- the type of TT fingerprint ('normal', 'hashed', 'bv')
      nBits -- the size of the bit vector (only for fpType='bv')
      minLength -- the minimum path length for an atom pair
      maxLength -- the maxmimum path length for an atom pair
      nBitsPerEntry -- the number of bits available for each torsion

    any additional keyword arguments will be passed to the fingerprinting function.

    """
  if fpType not in ['normal', 'hashed', 'bv']:
    raise ValueError("Unknown Topological torsion fingerprint type")
  if atomId < 0:
    return ttDict[fpType](mol, nBits, targetSize, nBitsPerEntry, 0, **kwargs)
  if atomId >= mol.GetNumAtoms():
    raise ValueError("atom index greater than number of atoms")
  return ttDict[fpType](mol, nBits, targetSize, nBitsPerEntry, [atomId], **kwargs)


# usage:   lambda m,i: GetMorganFingerprint(m, i, radius, fpType, nBits, useFeatures)
def GetMorganFingerprint(mol, atomId=-1, radius=2, fpType='bv', nBits=2048, useFeatures=False,
                         **kwargs):
  """
    Calculates the Morgan fingerprint with the environments of atomId removed.

    Parameters:
      mol -- the molecule of interest
      radius -- the maximum radius
      fpType -- the type of Morgan fingerprint: 'count' or 'bv'
      atomId -- the atom to remove the environments for (if -1, no environments is removed)
      nBits -- the size of the bit vector (only for fpType = 'bv')
      useFeatures -- if false: ConnectivityMorgan, if true: FeatureMorgan

    any additional keyword arguments will be passed to the fingerprinting function.
    """
  if fpType not in ['bv', 'count']:
    raise ValueError("Unknown Morgan fingerprint type")

  isBitVect = fpType == 'bv'
  if not hasattr(mol, '_fpInfo'):
    info = {}
    # get the fingerprint & construct the bitmap template
    if isBitVect:
      molFp = rdMD.GetMorganFingerprintAsBitVect(mol, radius, nBits=nBits, useFeatures=useFeatures,
                                                 bitInfo=info, **kwargs)
      bitmap = [DataStructs.ExplicitBitVect(nBits) for _ in range(mol.GetNumAtoms())]

    else:
      molFp = rdMD.GetMorganFingerprint(mol, radius, useFeatures=useFeatures, bitInfo=info,
                                        **kwargs)
      bitmap = [[] for _ in range(mol.GetNumAtoms())]

    # construct the bit map
    for bit, es in info.items():
      for at1, rad in es:
        if rad == 0:  # for radius 0
          if isBitVect:
            bitmap[at1][bit] = 1
          else:
            bitmap[at1].append(bit)
        else:  # for radii > 0
          env = Chem.FindAtomEnvironmentOfRadiusN(mol, rad, at1)
          amap = {}
          Chem.PathToSubmol(mol, env, atomMap=amap)
          for at2 in amap.keys():
            if isBitVect:
              bitmap[at2][bit] = 1
            else:
              bitmap[at2].append(bit)
    mol._fpInfo = (molFp, bitmap)

  if atomId < 0:
    return mol._fpInfo[0]
  # remove the bits of atomId
  if atomId >= mol.GetNumAtoms():
    raise ValueError("atom index greater than number of atoms")
  if len(mol._fpInfo) != 2:
    raise ValueError("_fpInfo not set")
  if isBitVect:
    molFp = mol._fpInfo[0] ^ mol._fpInfo[1][atomId]  # xor
  else:  # count
    molFp = copy.deepcopy(mol._fpInfo[0])
    # delete the bits with atomId
    for bit in mol._fpInfo[1][atomId]:
      molFp[bit] -= 1
  return molFp


# usage:   lambda m,i: GetRDKFingerprint(m, i, fpType, nBits, minPath, maxPath, nBitsPerHash)
def GetRDKFingerprint(mol, atomId=-1, fpType='bv', nBits=2048, minPath=1, maxPath=5, nBitsPerHash=2,
                      **kwargs):
  """
    Calculates the RDKit fingerprint with the paths of atomId removed.

    Parameters:
      mol -- the molecule of interest
      atomId -- the atom to remove the paths for (if -1, no path is removed)
      fpType -- the type of RDKit fingerprint: 'bv'
      nBits -- the size of the bit vector
      minPath -- minimum path length
      maxPath -- maximum path length
      nBitsPerHash -- number of to set per path
    """
  if fpType not in ['bv', '']:
    raise ValueError("Unknown RDKit fingerprint type")

  fpType = 'bv'
  if not hasattr(mol, '_fpInfo'):
    info = []  # list with bits for each atom
    # get the fingerprint
    molFp = Chem.RDKFingerprint(mol, fpSize=nBits, minPath=minPath, maxPath=maxPath,
                                nBitsPerHash=nBitsPerHash, atomBits=info, **kwargs)
    mol._fpInfo = (molFp, info)

  if atomId < 0:
    return mol._fpInfo[0]

  # remove the bits of atomId
  if atomId >= mol.GetNumAtoms():
    raise ValueError("atom index greater than number of atoms")
  if len(mol._fpInfo) != 2:
    raise ValueError("_fpInfo not set")

  molFp = copy.deepcopy(mol._fpInfo[0])
  molFp.UnSetBitsFromList(mol._fpInfo[1][atomId])
  return molFp



# Copyright (C) Brian Kelley 2021
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
"""
Free Wilson Analysis

A Mathematical Contribution to Structurre-Activity Studies
Spencer M, Free, Jr and James W Wilson
Journal of Medicinal Chemistry
Vol 7, Number 4, July 6, 1964

Basic usage: get a scaffold (or scaffolds) some compounds and their scores, 
then run freewilson:


Read some example data:
```
>>> import os, sys
>>> DATA_PATH = "data"
>>> smilesfile = os.path.join(DATA_PATH, "CHEMBL2321810.smi")
>>> scaffoldfile = os.path.join(DATA_PATH, "CHEMBL2321810_scaffold.mol")
>>> csvfile = os.path.join(DATA_PATH, "CHEMBL2321810_act.csv")
>>> mols = []
>>> for line in open(smilesfile):
...     smiles, name = line.strip().split()
...     m = Chem.MolFromSmiles(smiles)
...     m.SetProp("_Name", name)
...     mols.append(m)
>>> scaffold = Chem.MolFromMolBlock(open(scaffoldfile).read())
>>> data = {k:float(v) for k,v in list(csv.reader(open(csvfile)))[1:]}
>>> scores = [data[m.GetProp("_Name")] for m in mols]

```

And do the decomposition:
```
>>> from freewilson import FWDecompose, FWBuild, predictions_to_csv
>>> decomp = FWDecompose(scaffold, mols, scores)

```

Scores need to be in an appropriate form for regression analysis, i.e. pIC50s as opposed to IC50s.
Enumerations are a lot faster if you know a prediction cutoff or molecular weight cutoff.

Scaffolds can be any scaffold or smarts pattern or list of either.  The system automatically 
enumerates matching cores in the analysis.

To see if the decomposition is useful, check the r squared value

```
>>> print(f"Training R^2 is {decomp.r2:0.2f}")
Training R^2 is 0.81

```
Finally you can build the decomposition into new molecules:

```
>>> for pred in FWBuild(decomp):                            # doctest: +SKIP
...     print(pred.smiles, pred.prediction)

```

Now this builds both well and poorly predicted molecules.  To prevent
this you can use the following filters while building:

   1. pred_filter:  given a prediction, return True to keep the molecule
   2. hvy_filter: given a heavy atom count, return True to keep the molecule
   3. mw_filter: given a molecular weight, return True to keep the molecule
   4. mol_filter: given a sanitized molecule, return True to keep the molecule

Here are some examples (see using Molecular Filters below)

```
>>> preds = FWBuild(decomp, 
...                 pred_filter=lambda x: x > 8, 
...                 mw_filter=lambda mw: 100<mw<550)
>>> predictions_to_csv(sys.stdout, decomp, preds)

```

More Info
---------
You can also get some more information by setting logging to INFO

```
>>> import logging
>>> logging.getLogger().setLevel(logging.INFO)
>>> preds = list(FWBuild(decomp, pred_filter=lambda x: x > 8))

```

You'll see something like this:

```
INFO:root:Enumerating rgroups with no broken cycles...
INFO:root:	Core	1
INFO:root:	R1	73
INFO:root:	R10	2
INFO:root:	R3	445
100%|███████████████████████████████████████████████████████| 64970/64970 [00:05<00:00, 11247.38it/s]
INFO:root:Wrote 3 results out of 64970
	In Training set: 628
	Bad MW: 0
	Bad Pred: 64339
	Bad Filters: 0
	Bad smi: 0
	min mw: 380.429
	max mw: 772.4030000000001
	
	min pred: 2.8927324757327666
	max pred: 8.148182660251193
```

Using FMCS to find a scaffold
-----------------------------

If you don't have a good feel for the dataset, you can generate the scaffold by using
the rdFMCS package.  The following tries to find a decent MCS that covers 80% of the
input structures


```
>>> from rdkit.Chem import rdFMCS
>>> mcs = rdFMCS.FindMCS(mols[0:8], threshold=0.8, atomCompare=rdFMCS.AtomCompare.CompareAny,
...                      completeRingsOnly=True)                           # doctest: +SKIP
>>> decomp = FWDecompose(mcs.queryMol, mols, scores)                       # doctest: +SKIP

```
"""
import csv
import itertools
import logging
import math
import re
import sys
from collections import defaultdict, namedtuple
from typing import Generator, List

from sklearn.linear_model import Ridge
from sklearn.metrics import r2_score
from tqdm import tqdm

from rdkit import Chem, rdBase
from rdkit.Chem import Descriptors, molzip
from rdkit.Chem import rdRGroupDecomposition as rgd

logger = logging.getLogger("freewilson")

FreeWilsonPrediction = namedtuple("FreeWilsonPrediction", ['prediction', 'smiles', 'rgroups', 'mol', 'is_training'])

# match dummy atoms in a smiles string to extract atom maps
dummypat = re.compile(r"\*:([0-9]+)")


# molzip doesn't handle some of the forms that the RGroupDecomposition
#  returns, this solves these issues.

def molzip_smi(smiles):
  """Fix a rgroup smiles for molzip, note that the core MUST come first
    in the smiles string, ala core.rgroup1.rgroup2 ...
    """
  dupes = set()
  sl = []
  for s in smiles.split("."):
    if s.count("*") >= 1:
      if s in dupes:
        continue
      else:
        dupes.add(s)
    sl.append(s)

  smiles = ".".join(sl)
  m = Chem.RWMol(Chem.MolFromSmiles(smiles, sanitize=False))
  frags = Chem.GetMolFrags(m)
  core = frags[0]
  atommaps = {}
  counts = defaultdict(int)
  for idx in core:
    atommap = m.GetAtomWithIdx(idx).GetAtomMapNum()
    if atommap:
      atommaps[atommap] = idx
      counts[atommap] += 1

  next_atommap = max(atommaps) + 1
  add_atommap = []
  for fragment in frags[1:]:
    for idx in fragment:
      atommap = m.GetAtomWithIdx(idx).GetAtomMapNum()
      if atommap:
        count = counts[atommap] = counts[atommap] + 1
        if count > 2:
          m.GetAtomWithIdx(idx).SetAtomMapNum(next_atommap)
          add_atommap.append((atommaps[atommap], next_atommap))
          next_atommap += 1

  for atomidx, atommap in add_atommap:
    atom = m.GetAtomWithIdx(atomidx)
    bonds = list(atom.GetBonds())
    if len(bonds) == 1:
      oatom = bonds[0].GetOtherAtom(atom)
      xatom = Chem.Atom(0)
      idx = m.AddAtom(xatom)
      xatom = m.GetAtomWithIdx(idx)
      xatom.SetAtomMapNum(atommap)
      m.AddBond(oatom.GetIdx(), xatom.GetIdx(), Chem.BondType.SINGLE)
  return Chem.molzip(m)

def molzip_mols(mols):
  """Fix a rgroup smiles for molzip, note that the core MUST come first
    in the smiles string, ala core.rgroup1.rgroup2 ...
    """
  if not mols: return Chem.RWMol()
  
  m = Chem.RWMol(mols[0])
  dupes = set()
  for mol in mols[1:]:
    s = Chem.MolToSmiles(mol)
    if s.count("*") >= 1:
      if s in dupes:
        continue
      else:
        dupes.add(s)
    
    m.InsertMol(mol)

  frags = Chem.GetMolFrags(m)
  core = frags[0]
  atommaps = {}
  counts = defaultdict(int)
  for idx in core:
    atommap = m.GetAtomWithIdx(idx).GetAtomMapNum()
    if atommap:
      atommaps[atommap] = idx
      counts[atommap] += 1

  next_atommap = max(atommaps) + 1
  add_atommap = []
  for fragment in frags[1:]:
    for idx in fragment:
      atommap = m.GetAtomWithIdx(idx).GetAtomMapNum()
      if atommap:
        count = counts[atommap] = counts[atommap] + 1
        if count > 2:
          m.GetAtomWithIdx(idx).SetAtomMapNum(next_atommap)
          add_atommap.append((atommaps[atommap], next_atommap))
          next_atommap += 1

  for atomidx, atommap in add_atommap:
    atom = m.GetAtomWithIdx(atomidx)
    bonds = list(atom.GetBonds())
    if len(bonds) == 1:
      oatom = bonds[0].GetOtherAtom(atom)
      xatom = Chem.Atom(0)
      idx = m.AddAtom(xatom)
      xatom = m.GetAtomWithIdx(idx)
      xatom.SetAtomMapNum(atommap)
      m.AddBond(oatom.GetIdx(), xatom.GetIdx(), Chem.BondType.SINGLE)

  return Chem.molzip(m)


class RGroup:
  """FreeWilson RGroup
        smiles - isomeric smiles of the rgroup
        rgroup - Rgroup name ['Core', 'R1', 'R2'...]
        count - number of molecules with the rgroup
        coefficient - ridge regression coefficient
        idx - one-hot encoding for the rgroup
    """

  def __init__(self, smiles, rgroup, count, coefficient, idx=None, mol=None):
    self.smiles = smiles  # smiles for the sidechain (n.b. can be a core as well)
    self.rgroup = rgroup  # rgroup Core, R1, R2,...
    self.count = count  # num molecules with this rgruop
    self.coefficient = coefficient  # ridge coefficient
    self.idx = idx  # descriptor index
    self.dummies = tuple([int(x) for x in sorted(dummypat.findall(smiles))])
    self.mol = mol

    # Assemble some additive properties

    # will fail on some structures
    m = Chem.MolFromSmiles(smiles)
    if m:
      self.mw = Descriptors.MolWt(m)
      self.hvyct = Descriptors.HeavyAtomCount(m)
    else:
      # guess the MW if we can't sanitize
      table = Chem.GetPeriodicTable()
      m = Chem.MolFromSmiles(smiles, sanitize=False)
      if m:
        self.mw = 0.
        self.hvyct = 0
        for atom in m.GetAtoms():
          atomicnum = atom.GetAtomicNum()
          self.mw += table.GetAtomicWeight(atomicnum)
          if atomicnum > 1:
            self.hvyct += 1
        self.hvyct = Descriptors.HeavyAtomCount(m)
      else:
        self.mw = 0  # dunno
        self.hvyct = 0

  def __str__(self):
    return f"RGroup(smiles={repr(self.smiles)}, rgroup={repr(self.rgroup)}, count={repr(self.count)}, coefficient={repr(self.coefficient)}, idx={repr(self.idx)})"

  def __repr__(self):
    return self.__str__()


class FreeWilsonDecomposition:
  """FreeWilson decomposition
      rgroups - dictionary of rgroup to list of RGroups
                 i.e. {'Core': [RGroup(...), ...]
                       'R1': [ RGroup(...), RGroup(...)],
                      }
      rgroup_to_descriptor_idx - one hot encoding of smiles to descriptor index
      fitter - scikit learn compatible fitter used for regression
      N - number of rgroups
      r2 - regression r squared
      descriptors - set of the descriptors for molecules in the training set
                    used to not enumerate existing molecules
      row_decomposition - original rgroup decomposition (With row key 'molecule' is an rdkit molecule)
    """

  def __init__(self, rgroups, rgroup_to_descriptor_idx, fitter, r2, descriptors, row_decomposition,
               num_training, num_reconstructed):
    self.rgroups = rgroups  # dictionary 'Core':[core1, core1], 'R1': [rgroup1, rgroup2], ...
    self.rgroup_to_descriptor_idx = rgroup_to_descriptor_idx  # dictionary {smi:descriptor_idx}
    self.fitter = fitter  # fitter rgroup indices -> prediction
    self.N = len(rgroup_to_descriptor_idx)
    self.r2 = r2
    self.descriptors = set([tuple(d) for d in descriptors])
    self.row_decomposition = row_decomposition
    self.num_training = num_training
    self.num_reconstructed = num_reconstructed


default_decomp_params = rgd.RGroupDecompositionParameters()

# The default decomposition uses the GeneticAlgorirthm
#   fingerprint analysis to break symmetry and make
#   more consistent rgroup decompositions
default_decomp_params.matchingStrategy = rgd.GA

# Also the decomposition is allowed to add new rgroups
default_decomp_params.onlyMatchAtRGroups = False

# use the fingerprint variance method for scoring
default_decomp_params.scoreMethod = rgd.RGroupScore.FingerprintVariance

# we need to keep hydrogens so molzip will work
default_decomp_params.removeHydrogensPostMatch = False

class DecompEntry:
  def __init__(self, mol):
    self.mol = mol
    self.smiles = Chem.MolToSmiles(mol)
    
def FWDecompose(scaffolds, mols, scores,
                decomp_params=default_decomp_params) -> FreeWilsonDecomposition:
  """
    Perform a free wilson analysis
        : param scaffolds : scaffold or list of scaffolds to use for the rgroup decomposition
        : param mols : molecules to decompose
        : param scores : list of floating point numbers for the regression (
                             you may need convert these to their logs in some cases)
        : param decomp_params : RgroupDecompositionParams default [
                                    default_decomp_params = rdkit.Chem.rdRGroupDecomposition.RGroupDecompositionParameters()
                                    default_decomp_params.matchingStrategy = rgd.GA
                                    default_decomp_params.onlyMatchAtRGroups = False
                                   ]
                                If you only want to decompose on specific group locations
                                set onlyMatchAtRGroups to True


        >>> from rdkit import Chem
        >>> from freewilson import FWBuild, FWDecompose
        >>> from rdkit.Chem import Descriptors
        >>> scaffold = Chem.MolFromSmiles("c1cccnc1")
        >>> mols = [Chem.MolFromSmiles("c1cccnc1"+"C"*(i+1)) for i in range(100)]
        >>> scores = [Descriptors.MolLogP(m) for m in mols]
        >>> fw = FWDecompose(scaffold, mols, scores)
        >>> for pred in FWBuild(fw):
        ...   print(pred)

    For an easy way to report predictions see 

       >>> from freewilson import FWBuild, predictions_to_csv
       >>> import sys
       >>> predictions_to_csv(sys.stdout, FWBuild(fw))

   
    See FWBuild docs to see how to filter predictions, molecular weight or molecular properties.
    """
  descriptors = []  # list of descriptors, one per matched molecules
  #    descriptors are 1/0 if a sidechain is present
  matched_scores = []  # scores from the matching molecules
  rgroup_idx = {}  # rgroup index into descriptor { smiles: idx }
  rgroups = defaultdict(list)  # final list of rgrups/sidechains

  if len(mols) != len(scores):
    raise ValueError(
      f"The number of molecules must match the number of scores #mols {len(mols)} #scores {len(scores)}"
    )
  # decompose the rgroups
  logger.info(f"Decomposing {len(mols)} molecules...")
  decomposer = rgd.RGroupDecomposition(scaffolds, decomp_params)
  matched = []
  matched_indices = []
  for i, (mol, score) in enumerate(tqdm(zip(mols, scores))):
    if decomposer.Add(mol) >= 0:
      matched_scores.append(score)
      matched.append(mol)
      matched_indices.append(i)

  decomposer.Process()
  logger.info(f"Matched {len(matched_scores)} out of {len(mols)}")
  if not (matched_scores):
    logger.error("No scaffolds matched the input molecules")
    return

  #d = decomposer.GetRGroupsAsRows(asSmiles=True)
  d = decomposer.GetRGroupsAsRows()
  
  decomposition = [ {rg: DecompEntry(m) for rg, m in row.items()} for row in d]
  
  logger.info("Get unique rgroups...")
  blocker = rdBase.BlockLogs()
  rgroup_counts = defaultdict(int)
  num_reconstructed = 0

  for num_mols, (row, idx) in enumerate(zip(decomposition, matched_indices)):
    row_smiles = []
    for rgroup, de in row.items():
      row_smiles.append(de.smiles)
      rgroup_counts[de.smiles] += 1
      if de.smiles not in rgroup_idx:
        rgroup_idx[de.smiles] = len(rgroup_idx)
        rgroups[rgroup].append(RGroup(de.smiles, rgroup, 0, 0, mol=de.mol))
    row['original_idx'] = idx
    reconstructed = ".".join(row_smiles)
    try:
      blocker = rdBase.BlockLogs()
      mol = molzip_smi(reconstructed)
      num_reconstructed += 1
    except:
      logging.error("failed reconstructing %s, %s", Chem.MolToSmiles(matched[num_mols]),
                    reconstructed)

  logger.info(f"Descriptor size {len(rgroup_idx)}")
  logger.info(f"Reconstructed {num_reconstructed} out of {num_mols}")

  # get the descriptors list, one-hot encoding per rgroup
  if num_reconstructed == 0:
    logging.warning("Could only reconstruct %s out of %s training molecules", num_mols,
                    num_reconstructed)

  for mol, row in zip(matched, decomposition):
    row['molecule'] = mol
    descriptor = [0] * len(rgroup_idx)
    descriptors.append(descriptor)
    for k, de in row.items():
      if k == "original_idx" or k == 'molecule': continue
      if de.smiles in rgroup_idx:
        descriptor[rgroup_idx[de.smiles]] = 1

  assert len(descriptors) == len(
    matched_scores
  ), f"Number of descriptors({len(descriptors)}) doesn't match number of matcved scores({len(matched_scores)})"

  # Perform the Ridge Regression
  logger.info("Ridge Regressing...")
  lm = Ridge()
  lm.fit(descriptors, matched_scores)
  preds = lm.predict(descriptors)
  r2 = r2_score(matched_scores, preds)
  logger.info(f"R2 {r2}")
  logger.info(f"Intercept = {lm.intercept_:.2f}")

  for sidechains in rgroups.values():
    for rgroup in sidechains:
      rgroup.count = rgroup_counts[rgroup.smiles]
      rgroup.coefficient = lm.coef_[rgroup_idx[rgroup.smiles]]
      rgroup.idx = rgroup_idx[rgroup.smiles]

  return FreeWilsonDecomposition(rgroups, rgroup_idx, lm, r2, descriptors, decomposition, num_mols,
                                 num_reconstructed)


def _enumerate(rgroups, fw, mw_filter=None, hvy_filter=None, pred_filter=None, mol_filter=None,
               keep_training_set=False):
  N = fw.N
  fitter = fw.fitter
  num_products = 1
  for r in rgroups:
    num_products *= len(r)

  wrote = 0
  in_training_set = 0
  rejected_pred = 0
  rejected_mw = 0
  rejected_hvy = 0
  good_pred = 0
  rejected_filters = 0
  rejected_bad = 0
  min_mw = 10000000
  max_mw = 0
  min_hvy = 10000000
  max_hvy = 0
  max_pred = -1e10
  min_pred = 1e10
  delta = num_products // 10 or 1

  for i, groups in tqdm(enumerate(itertools.product(*rgroups)), total=num_products):
    if i and i % delta == 0:
      logging.debug(
        f"Wrote {wrote} results out of {num_products}\n\t\n\tIn Training Set{in_training_set}\n\tBad MW: {rejected_mw}\n\tBad Pred: {rejected_pred}\n\tBad Filters: {rejected_filters}\n\tBad smi: {rejected_bad}\n\tmin mw: {min_mw}\n\tmax mw: {max_mw}\n\t\n\tmin pred: {min_pred}\n\tmax pred: {max_pred}",
        file=sys.stderr)

    mw = 0
    hvy = 0
    descriptors = [0] * N
    for g in groups:
      descriptors[g.idx] = 1
      mw += g.mw
      hvy += g.hvyct

    if tuple(descriptors) in fw.descriptors:
      in_training_set += 1
      is_training = True
      if not keep_training_set:
        continue
    else:
      is_training = False  

    min_mw = min(min_mw, mw)
    max_mw = max(max_mw, mw)
    if mw_filter and not mw_filter(mw):
      rejected_mw += 1
      continue

    min_hvy = min(min_hvy, hvy)
    max_hvy = max(max_hvy, hvy)
    if hvy_filter and not hvy_filter(hvy):
      rejected_hvy += 1
      continue

    pred = fitter.predict([descriptors])[0]
    max_pred = max(pred, max_pred)
    min_pred = min(pred, min_pred)
    if pred_filter and not pred_filter(pred):
      rejected_pred += 1
      continue
    good_pred += 1

    mols = [g.mol for g in groups]
    if None in mols:
      smiles = set([g.smiles for g in groups])  # remove dupes
      smi = ".".join(set([g.smiles for g in groups]))
      try:
        mol = molzip_smi(smi)
      except:
        rejected_bad += 1
        continue
    else:
      try:
        mol = molzip_mols(mols)
      except:
        rejected_bad += 1
        continue

    rejected = False
    if mol_filter and not mol_filter(mol):
      rejected_filters += 1
      continue

    out_smi = Chem.MolToSmiles(mol)
    yield FreeWilsonPrediction(pred, out_smi, groups, mol, is_training)
    wrote += 1
  logging.info(
    f"Wrote {wrote} results out of {num_products}\n\tIn Training set: {in_training_set}\n\tBad MW: {rejected_mw}\n\tBad Pred: {rejected_pred}\n\tBad Filters: {rejected_filters}\n\tBad smi: {rejected_bad}\n\tmin mw: {min_mw}\n\tmax mw: {max_mw}\n\tBad HVY: {rejected_hvy}\n\tBad Pred: {rejected_pred}\n\tBad Filters: {rejected_filters}\n\tBad smi: {rejected_bad}\n\tmin mw: {min_mw}\n\tmax mw: {max_mw}\n\tmin hvy: {min_hvy}\n\tmax hvy: {max_hvy}\n\t\n\tmin pred: {min_pred}\n\tmax pred: {max_pred}"
  )


def FWBuild(fw: FreeWilsonDecomposition, pred_filter=None, mw_filter=None, hvy_filter=None,
            mol_filter=None, keep_training_set=False) -> Generator[FreeWilsonPrediction, None, None]:
  """Enumerate the freewilson decomposition and return their predictions

       :param fw: FreeWilsonDecomposition generated from FWDecompose
       :param pred_filter: return True if the prediction is in a desireable range
                           e.g.  lambda pic50: pic50>=8
       :param mw_filter: return True if the enumerated molecular weight is in a desireable rrange
                           e.g. lambda mw: 150 < mw < 550
       :param hvy_filter: return True if the enumerated heavy couont is in a desireable rrange
                           e.g. lambda hvy: 10 < hvy < 50
       :param mol_filter: return True if the molecule is ok to be enumerated
                           e.g. lambda mol: -3 < Descriptors.MolLogp(mol) < 5
    """
  blocker = rdBase.BlockLogs()
  # check groups for cycles
  cycles = set()
  rgroups_no_cycles = defaultdict(list)
  rgroup_cycles = defaultdict(list)

  # we can handle cycles now?
  for key, rgroup in fw.rgroups.items():
    if key == 'Core':
      rgroups_no_cycles[key] = rgroup
      continue
    no_cycles = rgroups_no_cycles[key]
    for g in rgroup:
      no_cycles.append(g)
      continue

      if len(g.dummies) > 1:
        cycles.add(g.dummies)
        rgroup_cycles[g.dummies].append(g)
      else:
        no_cycles.append(g)

  logging.info("Enumerating rgroups with no broken cycles...")
  for k, v in rgroups_no_cycles.items():
    logging.info(f"\t{k}\t{len(v)}")
  # do the ones that have no cycles first
  rgroups = [rgroup for key, rgroup in sorted(rgroups_no_cycles.items())]
  # core is always first
  for res in _enumerate(rgroups, fw, pred_filter=pred_filter, mw_filter=mw_filter,
                        hvy_filter=hvy_filter, mol_filter=mol_filter,
                        keep_training_set=keep_training_set):
    yield res

  # iterate on rgroups with cycles
  #  basically only let one set of RGroups show up once.
  indices = set()
  for k in fw.rgroups:
    if k[0] == "R":
      indices.add(int(k[1:]))
  if cycles:
    logging.info("Enumerating rgroups with broken cycles...")

  for rgroup_indices in cycles:
    missing = indices - set(rgroup_indices)
    rgroups = {'Core': fw.rgroups['Core']}
    rgroups["R%s" % ".".join([str(x) for x in rgroup_indices])] = rgroup_cycles[rgroup_indices]
    for m in missing:
      k = "R%s" % m
      rgroups[k] = rgroups_no_cycles[k]

    for k, v in rgroups.items():
      logging.info(f"\t{k}\t{len(v)}")

    for res in _enumerate([rgroup for key, rgroup in sorted(rgroups.items())], fw,
                          pred_filter=pred_filter, mw_filter=mw_filter, hvy_filter=hvy_filter,
                          mol_filter=mol_filter):
      yield res


def _rgroup_sort(r):
  """Sort groups like R1 R2 R10 not R1 R10 R2
    """
  if r[0] == "R":
    return ("R", int(r[1:]))
  return (r, None)


def predictions_to_csv(outstream, decomposition: FreeWilsonDecomposition, predictions):
  """Output predictions in csv format to the output stream

       :param outstream: output stream to write results
       :param decomposition: freewillson decomposition
       :param predictions: list of Predictions to output
    """
  writer = None
  for pred in predictions:
    if not writer:
      rgroups = set()
      for rgroup in decomposition.rgroups:
        rgroups.add(rgroup)
      rgroups = sorted(rgroups, key=_rgroup_sort)

      lookup = {}
      for i, rg in enumerate(rgroups):
        lookup[rg] = i
      writer = csv.writer(outstream)
      header = ['smiles', 'prediction'] + [f"{rg}_smiles" for rg in list(rgroups)]
      writer.writerow(header)
    rg = [""] * len(lookup)
    for s in pred.rgroups:
      rg[lookup[s.rgroup]] = s.smiles

    row = [pred.smiles, repr(pred.prediction)] + rg
    writer.writerow(row)
  return header


def test_freewilson():
  # some simple tests
  from rdkit import Chem
  from rdkit.Chem import Descriptors
  assert dummypat.findall("C[*:1]N.[H][*:2]") == ['1', '2']
  assert dummypat.findall("C[*:1]N.[HH][*:2]") == ['1', '2']
  assert dummypat.findall("C[*:1]N.[2H][*:2]") == ['1', '2']
  assert dummypat.findall("C[*:1]N.[CH2][*:2]") == ['1', '2']

  scaffold = Chem.MolFromSmiles("[*:2]c1cccnc1[*:1]")
  mols = [Chem.MolFromSmiles("N" * (i + 1) + "c1cccnc1" + "C" * (i + 1)) for i in range(10)]
  scores = [Descriptors.MolLogP(m) for m in mols]
  fw = FWDecompose(scaffold, mols, scores)

  for pred in FWBuild(fw, pred_filter=lambda x: -3 < x < 3, mw_filter=lambda mw: 100 < mw < 450,
                      hvy_filter=lambda hvy: 10 < hvy < 50,
                      mol_filter=lambda m: -3 < Descriptors.MolLogP(m) < 3):
    rgroups = set()
    for sidechain in pred.rgroups:
      rgroups.add(sidechain.rgroup)
    rgroups = sorted(rgroups)
    assert list(rgroups) == ['Core', 'R1', 'R2']


def test_rgroups():
  smiles = "[*:2]c1ccccc1[*:1]"
  rg = RGroup(smiles=smiles, rgroup="Core", count=1, coefficient=0.1)
  mol = Chem.MolFromSmiles(smiles)
  assert rg.mw == Descriptors.MolWt(mol)
  assert rg.hvyct == Descriptors.HeavyAtomCount(mol)
  assert rg.dummies == (1, 2)

  # try a non parseable smiles
  smiles = "[*:2]cccccc[*:1]"
  rg = RGroup(smiles=smiles, rgroup="Core", count=1, coefficient=0.1)
  assert rg.mw == 72.06599999999999  # almost but not quite benzene
  assert rg.hvyct == 6
  assert rg.dummies == (1, 2)

  # finally aa non parseble one
  smiles = "Nope"
  rg = RGroup(smiles=smiles, rgroup="Core", count=1, coefficient=0.1)
  assert rg.mw == 0
  assert rg.hvyct == 0
  assert rg.dummies == tuple()



#
#  Copyright (C) 2011-2021 Greg Landrum and other RDKit contributors
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
import base64
import copy
import html
import warnings
from io import BytesIO

import IPython
from IPython.display import HTML, SVG

from rdkit import Chem
from rdkit.Chem import Draw, rdchem, rdChemReactions
from rdkit.Chem.Draw import rdMolDraw2D

from . import InteractiveRenderer

if IPython.release.version < '0.11':
  raise ImportError('this module requires at least v0.11 of IPython')
try:
  import py3Dmol
  _canUse3D = True
except ImportError:
  _canUse3D = False

from PIL import Image
from PIL.PngImagePlugin import PngInfo

molSize = (450, 150)
highlightSubstructs = True  # highlight substructure matches when drawing 2D structures
kekulizeStructures = True  # try to kekulize structures before drawing
highlightByReactant = False  # highlight reactions by reactant
ipython_useSVG = False  # use the SVG renderer for 2D depictions (otherwise PNG is used)
ipython_showProperties = True  # display molecule properties when rendering them in 2D
ipython_maxProperties = 10  # maximum number of properties to display
ipython_3d = False  # try to use 3D rendering for molecules with a 3D conformer
molSize_3d = (400, 400)  # default size of 3D structures
drawing_type_3d = 'stick'  # default drawing type for 3d structures
bgcolor_3d = '0xeeeeee'  # default background color for 3d structures
drawOptions = rdMolDraw2D.MolDrawOptions()  # drawing options for 2D structures
InteractiveRenderer._defaultDrawOptions = drawOptions


def addMolToView(mol, view, confId=-1, drawAs=None):
  ''' adds a single molecule to a py3Dmol view

Parameters
----------
m : RDKit molecule
    the molecule to draw
view : py3Dmol.view
    the view to draw into
confId : int
    the conformer ID to draw. If not provided, the default conformer (-1) will be used
drawAs : str
    the drawing type to use. If not provided, 'stick' will be used for small molecules, 
    'cartoon' for large (more than 999 atoms) molecules
    
Returns
-------
None'''
  if mol.GetNumAtoms() >= 999 or drawAs == 'cartoon':
    # py3DMol is happier with TER and MASTER records present
    pdb = Chem.MolToPDBBlock(mol, flavor=0x20 | 0x10)
    view.addModel(pdb, 'pdb')
  else:
    # py3Dmol does not currently support v3k mol files, so
    # we can only provide those with "smaller" molecules
    try:
      mb = Chem.MolToMolBlock(mol, confId=confId)
    except Chem.AtomKekulizeException:
      # The bonding is likely to be wrong, but it would be good
      # to see something.
      p = Chem.rdmolfiles.MolWriterParams()
      p.kekulize = False
      mb = Chem.MolToMolBlock(mol, confId=confId, params=p)

    view.addModel(mb, 'sdf')
  if drawAs is None:
    drawAs = drawing_type_3d
  view.setStyle({drawAs: {}})


def drawMol3D(m, view=None, confId=-1, drawAs=None, bgColor=None, size=None):
  ''' draws a single molecule in 3D using py3Dmol

Parameters
----------
m : RDKit molecule
    the molecule to draw
view : py3Dmol.view
    the view to draw into. If not provided, one will be created
confId : int
    the conformer ID to draw. If not provided, the default conformer (-1) will be used
drawAs : str
    the drawing type to use. If not provided, 'stick' will be used
bgColor : str
    the background color to use for the drawing. If not provided, the default background color will be used
size : tuple of int
    the size of the drawing. If not provided, molSize_3d will be used

Returns
-------
the py3Dmol.view object containing the drawing'''
  if bgColor is None:
    bgColor = bgcolor_3d
  if size is None:
    size = molSize_3d
  if view is None:
    view = py3Dmol.view(width=size[0], height=size[1])
  view.removeAllModels()

  try:
    ms = iter(m)
    for m in ms:
      addMolToView(m, view, confId, drawAs)
  except TypeError:
    addMolToView(m, view, confId, drawAs)
  view.setBackgroundColor(bgColor)
  view.zoomTo()
  return view.show()


def drawMols3D(mols, view=None, confIds=None, drawAs=None, bgColor=None, size=None, removeHs=False,
               colors=('cyanCarbon', 'redCarbon', 'blueCarbon', 'magentaCarbon', 'yellowCarbon',
                       'cornflowerblueCarbon')):
  ''' draws a list/tuple of molecules in 3D using py3Dmol

Parameters
----------
mols : sequence (list, tuple, etc.) of RDKit molecules
    the molecules to draw
view : py3Dmol.view
    the view to draw into. If not provided, one will be created
confIds : sequence (list, tuple, etc.) of int
    the conformer ID to draw for each molecule. If not provided, the default conformer (-1) will be used
drawAs : str or sequence of str
    the drawing type to use for each molecule. If not provided, 'stick' will be used
bgColor : str
    the background color to use for the drawing. If not provided, the default background color will be used
size : tuple of int
    the size of the drawing. If not provided, molSize_3d will be used
removeHs : bool
    whether or not to remove Hs from the molecules before drawing, the default is False
colors : sequence of str
    the colors to use for drawing the molecules. 

Returns
-------
the py3Dmol.view object containing the drawing'''
  if bgColor is None:
    bgColor = bgcolor_3d
  if size is None:
    size = molSize_3d
  if view is None:
    view = py3Dmol.view(width=size[0], height=size[1])

  if drawAs is None:
    drawAs = ['stick'] * len(mols)
  if confIds is None:
    confIds = [-1] * len(mols)

  for m, confId in zip(mols, confIds):
    if removeHs:
      m = Chem.RemoveHs(m)
    addMolToView(m, view, confId)
  for i in range(len(mols)):
    style = {'colorscheme': colors[i % len(colors)]}
    if drawAs[i] == "sphere":
      style["radius"] = 0.5
    view.setStyle({
      'model': i,
    }, {drawAs[i]: style})

  view.setBackgroundColor(bgColor)
  view.zoomTo()
  return view.show()


def _toJSON(mol):
  """For IPython notebook, renders 3D webGL objects."""
  if not ipython_3d or not mol.GetNumConformers():
    return None
  conf = mol.GetConformer()
  if not conf.Is3D():
    return None
  res = drawMol3D(mol)
  if hasattr(res, 'data'):
    return res.data
  return ""


def _wrapHTMLIntoTable(html):
  return InteractiveRenderer.injectHTMLFooterAfterTable(
    f'<div><table><tbody><tr><td style="width: {molSize[0]}px; ' +
    f'height: {molSize[1]}px; text-align: center;">' + html.replace(" scoped", "") +
    '</td></tr></tbody></table></div>')


def _toHTML(mol):
  useInteractiveRenderer = InteractiveRenderer.isEnabled(mol)
  if _canUse3D and ipython_3d and mol.GetNumConformers() and mol.GetConformer().Is3D():
    return _toJSON(mol)
  props = mol.GetPropsAsDict()
  if not ipython_showProperties or not props:
    if useInteractiveRenderer:
      return _wrapHTMLIntoTable(
        InteractiveRenderer.generateHTMLBody(mol, molSize, useSVG=ipython_useSVG))
    else:
      return _toSVG(mol)
  if mol.HasProp('_Name'):
    nm = mol.GetProp('_Name')
  else:
    nm = ''

  res = []

  if useInteractiveRenderer:
    content = InteractiveRenderer.generateHTMLBody(mol, molSize, legend=nm, useSVG=ipython_useSVG)
  else:
    if not ipython_useSVG:
      png = _toPNG(mol)
      png = base64.b64encode(png)
      content = f'<image src="data:image/png;base64,{png.decode()}">'
    else:
      content = _toSVG(mol)
  res.append(f'<tr><td colspan="2" style="text-align: center;">{content}</td></tr>')

  for i, (pn, pv) in enumerate(props.items()):
    if ipython_maxProperties >= 0 and i >= ipython_maxProperties:
      res.append(
        '<tr><td colspan="2" style="text-align: center">Property list truncated.<br />Increase IPythonConsole.ipython_maxProperties (or set it to -1) to see more properties.</td></tr>'
      )
      break
    pv = html.escape(str(pv))
    res.append(
      f'<tr><th style="text-align: right">{pn}</th><td style="text-align: left">{pv}</td></tr>')
  res = '\n'.join(res)
  res = f'<table>{res}</table>'
  if useInteractiveRenderer:
    res = InteractiveRenderer.injectHTMLFooterAfterTable(res)
  return res


def listToLists(lst):
  if not lst or not isinstance(lst, (list, tuple)):
    return []
  if isinstance(lst[0], (list, tuple)):
    return lst
  return [lst]


def _toPNG(mol):
  if hasattr(mol, '__sssAtoms'):
    highlightAtoms = listToLists(mol.__sssAtoms)
    return Draw.DrawMolWithMatches(mol, highlightAtoms, molSize=molSize,
                                   qry=getattr(mol, '__sssQry', None), label='',
                                   options=drawOptions, doPNG=True, kekulize=kekulizeStructures)
  else:
    highlightAtoms = []
  kekulize = kekulizeStructures
  return Draw._moltoimg(mol, molSize, highlightAtoms, "", returnPNG=True, kekulize=kekulize,
                        drawOptions=drawOptions)


def _toSVG(mol):
  if not ipython_useSVG:
    return None
  if hasattr(mol, '__sssAtoms'):
    highlightAtoms = listToLists(mol.__sssAtoms)
    return Draw.DrawMolWithMatches(mol, highlightAtoms, molSize=molSize,
                                   qry=getattr(mol, '__sssQry', None), label='',
                                   options=drawOptions, doPNG=False, kekulize=kekulizeStructures)
  else:
    highlightAtoms = []
  kekulize = kekulizeStructures
  return Draw._moltoSVG(mol, molSize, highlightAtoms, "", kekulize, drawOptions=drawOptions)


def _toReactionPNG(rxn):
  rc = copy.deepcopy(rxn)
  return Draw.ReactionToImage(rc, subImgSize=(int(molSize[0] / 3), molSize[1]),
                              highlightByReactant=highlightByReactant, drawOptions=drawOptions,
                              returnPNG=True)


def _toReactionSVG(rxn):
  if not ipython_useSVG:
    return None
  rc = copy.deepcopy(rxn)
  return Draw.ReactionToImage(rc, subImgSize=(int(molSize[0] / 3), molSize[1]), useSVG=True,
                              highlightByReactant=highlightByReactant, drawOptions=drawOptions)


def _toMolBundlePNG(bundle):
  if Draw._MolsToGridImageSaved is not None:
    fn = Draw._MolsToGridImageSaved
  else:
    fn = Draw.MolsToGridImage
  return fn(bundle, subImgSize=molSize, drawOptions=drawOptions, useSVG=False, returnPNG=True)


def _toMolBundleSVG(bundle):
  if not ipython_useSVG:
    return None
  if Draw._MolsToGridImageSaved is not None:
    fn = Draw._MolsToGridImageSaved
  else:
    fn = Draw.MolsToGridImage
  return fn(bundle, subImgSize=molSize, drawOptions=drawOptions, useSVG=True)


def _GetSubstructMatch(mol, query, *args, **kwargs):
  res = mol.__GetSubstructMatch(query, *args, **kwargs)
  if highlightSubstructs:
    mol.__sssAtoms = list(res)
    mol.__sssQry = query
  else:
    mol.__sssAtoms = []
    mol.__sssQry = None
  return res


_GetSubstructMatch.__doc__ = rdchem.Mol.GetSubstructMatch.__doc__


def _GetSubstructMatches(mol, query, *args, **kwargs):
  res = mol.__GetSubstructMatches(query, *args, **kwargs)
  mol.__sssAtoms = []
  if highlightSubstructs:
    mol.__sssQry = query
    mol.__sssAtoms = res
  else:
    mol.__sssQry = None
  return res


_GetSubstructMatches.__doc__ = rdchem.Mol.GetSubstructMatches.__doc__


# code for displaying PIL images directly,
def display_pil_image(img):
  """displayhook function for PIL Images, rendered as PNG"""
  # pull metadata from the image, if there
  metadata = PngInfo()
  for k, v in img.info.items():
    metadata.add_text(k, v)
  bio = BytesIO()
  img.save(bio, format='PNG', pnginfo=metadata)
  return bio.getvalue()


from IPython import display


def ShowMols(mols, maxMols=50, **kwargs):
  if 'useSVG' not in kwargs:
    kwargs['useSVG'] = ipython_useSVG
  if 'returnPNG' not in kwargs:
    kwargs['returnPNG'] = True
  if InteractiveRenderer.isEnabled():
    fn = InteractiveRenderer.MolsToHTMLTable
  elif Draw._MolsToGridImageSaved is not None:
    fn = Draw._MolsToGridImageSaved
  else:
    fn = Draw.MolsToGridImage

  if len(mols) > maxMols:
    warnings.warn(
      "Truncating the list of molecules to be displayed to %d. Change the maxMols value to display more."
      % (maxMols))
    mols = mols[:maxMols]
    for prop in ('legends', 'highlightAtoms', 'highlightBonds'):
      if prop in kwargs:
        kwargs[prop] = kwargs[prop][:maxMols]
  if "drawOptions" not in kwargs:
    kwargs["drawOptions"] = drawOptions

  res = fn(mols, **kwargs)
  if InteractiveRenderer.isEnabled():
    return HTML(res)
  elif kwargs['useSVG']:
    return SVG(res)
  elif kwargs['returnPNG']:
    return display.Image(data=res, format='png')
  return res


def _DrawBit(fn, *args, **kwargs):
  if 'useSVG' not in kwargs:
    kwargs['useSVG'] = ipython_useSVG

  res = fn(*args, **kwargs)
  if kwargs['useSVG']:
    return SVG(res)
  sio = BytesIO(res)
  return Image.open(sio)


def _DrawBits(fn, *args, **kwargs):
  if 'useSVG' not in kwargs:
    kwargs['useSVG'] = ipython_useSVG

  res = fn(*args, **kwargs)
  if kwargs['useSVG']:
    return SVG(res)
  sio = BytesIO(res)
  return Image.open(sio)


def DrawMorganBit(mol, bitId, bitInfo, drawOptions=drawOptions, **kwargs):
  if Draw._DrawMorganBitSaved is not None:
    fn = Draw._DrawMorganBitSaved
  else:
    fn = Draw.DrawMorganBit
  return _DrawBit(fn, mol, bitId, bitInfo, drawOptions=drawOptions, **kwargs)


def DrawMorganBits(*args, drawOptions=drawOptions, **kwargs):
  if Draw._DrawMorganBitsSaved is not None:
    fn = Draw._DrawMorganBitsSaved
  else:
    fn = Draw.DrawMorganBits
  return _DrawBit(fn, *args, drawOptions=drawOptions, **kwargs)


def DrawRDKitBit(mol, bitId, bitInfo, drawOptions=drawOptions, **kwargs):
  if Draw._DrawRDKitBitSaved is not None:
    fn = Draw._DrawRDKitBitSaved
  else:
    fn = Draw.DrawRDKitBit
  return _DrawBit(fn, mol, bitId, bitInfo, drawOptions=drawOptions, **kwargs)


def DrawRDKitBits(*args, drawOptions=drawOptions, **kwargs):
  if Draw._DrawRDKitBitsSaved is not None:
    fn = Draw._DrawRDKitBitsSaved
  else:
    fn = Draw.DrawRDKitBits
  return _DrawBit(fn, *args, drawOptions=drawOptions, **kwargs)


def EnableSubstructMatchRendering():
  if not hasattr(rdchem.Mol, '__GetSubstructMatch'):
    rdchem.Mol.__GetSubstructMatch = rdchem.Mol.GetSubstructMatch
  rdchem.Mol.GetSubstructMatch = _GetSubstructMatch
  if not hasattr(rdchem.Mol, '__GetSubstructMatches'):
    rdchem.Mol.__GetSubstructMatches = rdchem.Mol.GetSubstructMatches
  rdchem.Mol.GetSubstructMatches = _GetSubstructMatches


_methodsToDelete = []
_rendererInstalled = False


def InstallIPythonRenderer():
  global _rendererInstalled
  if _rendererInstalled:
    return

  # in case of a reload, there may be some cleanup necessary in other modules that we have
  # monkey patched. So go ahead and do an uninstall to cleanup before we start. This was
  # Github #8082
  UninstallIPythonRenderer()

  rdchem.Mol._repr_png_ = _toPNG
  rdchem.Mol._repr_svg_ = _toSVG
  _methodsToDelete.append((rdchem.Mol, '_repr_png_'))
  _methodsToDelete.append((rdchem.Mol, '_repr_svg_'))
  rdchem.Mol._repr_html_ = _toHTML
  _methodsToDelete.append((rdchem.Mol, '_repr_html_'))

  rdChemReactions.ChemicalReaction._repr_png_ = _toReactionPNG
  rdChemReactions.ChemicalReaction._repr_svg_ = _toReactionSVG
  _methodsToDelete.append((rdChemReactions.ChemicalReaction, '_repr_png_'))
  _methodsToDelete.append((rdChemReactions.ChemicalReaction, '_repr_svg_'))

  rdchem.MolBundle._repr_png_ = _toMolBundlePNG
  rdchem.MolBundle._repr_svg_ = _toMolBundleSVG
  _methodsToDelete.append((rdchem.MolBundle, '_repr_png_'))
  _methodsToDelete.append((rdchem.MolBundle, '_repr_svg_'))

  EnableSubstructMatchRendering()
  Image.Image._repr_png_ = display_pil_image
  _methodsToDelete.append((Image.Image, '_repr_png_'))
  Draw._MolsToGridImageSaved = Draw.MolsToGridImage
  Draw.MolsToGridImage = ShowMols
  Draw.MolsToGridImage.__doc__ = Draw._MolsToGridImageSaved.__doc__
  Draw._DrawRDKitBitSaved = Draw.DrawRDKitBit
  Draw.DrawRDKitBit = DrawRDKitBit
  Draw.DrawRDKitBit.__doc__ = Draw._DrawRDKitBitSaved.__doc__
  Draw._DrawRDKitBitsSaved = Draw.DrawRDKitBits
  Draw.DrawRDKitBits = DrawRDKitBits
  Draw.DrawRDKitBits.__doc__ = Draw._DrawRDKitBitsSaved.__doc__
  Draw._DrawMorganBitSaved = Draw.DrawMorganBit
  Draw.DrawMorganBit = DrawMorganBit
  Draw.DrawMorganBit.__doc__ = Draw.DrawMorganBit.__doc__
  Draw._DrawMorganBitsSaved = Draw.DrawMorganBits
  Draw.DrawMorganBits = DrawMorganBits
  Draw.DrawMorganBits.__doc__ = Draw._DrawMorganBitsSaved.__doc__
  if not hasattr(rdchem.Mol, '__DebugMol'):
    rdchem.Mol.__DebugMol = rdchem.Mol.Debug
    rdchem.Mol.Debug = lambda self, useStdout=False: self.__DebugMol(useStdout=useStdout)

  _rendererInstalled = True


def DisableSubstructMatchRendering():
  if hasattr(rdchem.Mol, '__GetSubstructMatch'):
    rdchem.Mol.GetSubstructMatch = rdchem.Mol.__GetSubstructMatch
    del rdchem.Mol.__GetSubstructMatch
  if hasattr(rdchem.Mol, '__GetSubstructMatches'):
    rdchem.Mol.GetSubstructMatches = rdchem.Mol.__GetSubstructMatches
    del rdchem.Mol.__GetSubstructMatches


def UninstallIPythonRenderer():
  global _rendererInstalled, _methodsToDelete

  for cls, attr in _methodsToDelete:
    if hasattr(cls, attr):
      delattr(cls, attr)

  _methodsToDelete = []

  DisableSubstructMatchRendering()

  if hasattr(Draw, '_MolsToGridImageSaved') and Draw._MolsToGridImageSaved is not None:
    Draw.MolsToGridImage = Draw._MolsToGridImageSaved
    del Draw._MolsToGridImageSaved
  if hasattr(Draw, '_DrawRDKitBitSaved') and Draw._DrawRDKitBitSaved is not None:
    Draw.DrawRDKitBit = Draw._DrawRDKitBitSaved
    del Draw._DrawRDKitBitSaved
  if hasattr(Draw, '_DrawRDKitBitsSaved') and Draw._DrawRDKitBitsSaved is not None:
    Draw.DrawRDKitBits = Draw._DrawRDKitBitsSaved
    del Draw._DrawRDKitBitsSaved
  if hasattr(Draw, '_DrawMorganBitSaved') and Draw._DrawMorganBitSaved is not None:
    Draw.DrawMorganBit = Draw._DrawMorganBitSaved
    del Draw._DrawMorganBitSaved
  if hasattr(Draw, '_DrawMorganBitsSaved') and Draw._DrawMorganBitsSaved is not None:
    Draw.DrawMorganBits = Draw._DrawMorganBitsSaved
    del Draw._DrawMorganBitsSaved

  if hasattr(rdchem.Mol, '__DebugMol'):
    rdchem.Mol.Debug = rdchem.Mol.__DebugMol
    del rdchem.Mol.__DebugMol

  _rendererInstalled = False


InstallIPythonRenderer()



import sys
import os
import argparse
import pybind11_stubgen
from pybind11_stubgen.parser.mixins.parse import ExtractSignaturesFromPybind11Docstrings
from pybind11_stubgen.printer import Printer
import importlib


"""
worker script
1st param is the temporary directory
2nd param is the module name
"""
def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--tempdir",
                        help=f"Temporary directory")
    parser.add_argument("--module-name",
                        help=f"Python module name for which stubs will be generated")
    parser.add_argument("--outer-dirs", nargs="*",
                        help=f"Outer dirs where stubs should be copied")
    parser.add_argument("--keep-incorrect-staticmethods",
                        help=f"Whether incorrectly assigned staticmethods should be kept as such",
                        action="store_true")
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = parse_args()
    keep_incorrect_staticmethods = args.keep_incorrect_staticmethods or os.environ.get("KEEP_INCORRECT_STATICMETHODS", None)
    if isinstance(keep_incorrect_staticmethods, str):
        try:
            keep_incorrect_staticmethods = bool(int(keep_incorrect_staticmethods))
        except ValueError:
            keep_incorrect_staticmethods = False
    sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))
    gen_rdkit_stubs = importlib.import_module("gen_rdkit_stubs")

    if hasattr(ExtractSignaturesFromPybind11Docstrings, "parse_function_docstring"):
        parse_function_docstring_orig = ExtractSignaturesFromPybind11Docstrings.parse_function_docstring

        def parse_function_docstring_patched(self, func_name, doc_lines, **kwargs):
            doc_lines = gen_rdkit_stubs.ProcessDocLines.process(args.module_name, doc_lines, keep_incorrect_staticmethods)
            return parse_function_docstring_orig(self, func_name, doc_lines, **kwargs)

        ExtractSignaturesFromPybind11Docstrings.parse_function_docstring = parse_function_docstring_patched

    if hasattr(Printer, "print_submodule_import"):
        print_submodule_import_orig = Printer.print_submodule_import

        def print_submodule_import_patched(self, name):
            return [f"from .{name} import *"]

        Printer.print_submodule_import = print_submodule_import_patched


    stored_argv = list(sys.argv)
    try:
        sys.argv = ["",
                    "--root-suffix",
                    "",
                    "--ignore-all-errors",
                    "-o",
                    args.tempdir,
                    args.module_name]
        pybind11_stubgen.main()
    except Exception as e:
        if isinstance(e, AssertionError) or isinstance(e, ImportError):
            raise
        else:
            print(str(e))
    finally:
        sys.argv = stored_argv
    src_path = os.path.join(args.tempdir, *args.module_name.split("."))
    gen_rdkit_stubs.patch_stubs(args.tempdir, src_path)
    gen_rdkit_stubs.copy_stubs(src_path, args.outer_dirs)



#
#  Copyright (c) 2023, RDKit Hackathon 2023, implemented by:
# - Arsenio Cruz
# - José-Manuel Gally
# - Axel Pahl
# - Vincenzo Palmacci
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of Novartis Institutes for BioMedical Research Inc.
#     nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""
Spacial score (SPS) is an empirical scoring system to express the spacial complexity of a compound
in an uniform manner and on a highly granular scale for ranking and comparison between molecules. [1]
SPS takes into account the fraction of sp3 hybridized carbons and the fraction of stereogenic carbons.

By default, this module generates the normalized spacial score (nSPS), which is a variation of the SPS score
that considers the size of the molecule.
To obtain the nSPS score the SPS score is divided by the total number of heavy atoms in the analyzed molecule.

SPS = sum(h*s*r*n*n)
nSPS = SPS/a

Where:
h = Atom hybridisation term
s = Stereoisomeric term
r = Non-aromatic ring term
n = Number of heavy atom neighbors
a = Total number of heavy atoms in the molecule

The SPS function in this module takes a mol object and returns either the absolute score (normalize=False) or the score normalized by the number of heavy atoms (normalize=True (default)).

The original code implementation can be found at: https://github.com/frog2000/Spacial-Score/blob/main/spacial_score.py

[1] Krzyzanowski, A.; Pahl, A.; Grigalunas, M.; Waldmann, H. Spacial Score─A Comprehensive Topological Indicator for Small-Molecule Complexity. J. Med. Chem. 2023. https://doi.org/10.1021/acs.jmedchem.3c00689.
"""
from rdkit import Chem

# import rdkit.Chem.Descriptors as Desc
from rdkit.Chem import rdmolops
from rdkit.Chem.ChemUtils.DescriptorUtilities import setDescriptorVersion
from collections import defaultdict


@setDescriptorVersion(version="1.0.0")
def SPS(mol, normalize=True):
  """Calculates the SpacialScore descriptor. By default, the score is normalized by the number of heavy atoms (nSPS) resulting in a float value,
    otherwise (normalize=False) the absolute score is returned as an integer.
    """
  return _SpacialScore(mol, normalize=normalize).score


class _SpacialScore:
  """Class intended for calculating spacial score (SPS) and size-normalised SPS (nSPS) for small organic molecules"""

  def __init__(self, mol, normalize=True):
    if mol is None:
      raise ValueError("No valid molecule object found.")
    molCp = Chem.Mol(mol)
    rdmolops.FindPotentialStereoBonds(molCp)
    self.mol = molCp  # mol is supposed to be a valid RDKit Mol object
    self.normalize = normalize  # if true nSPS, otherwise SPS
    self.hyb_score = {}
    self.stereo_score = {}
    self.ring_score = {}
    self.bond_score = {}
    self.chiral_idxs = self.findStereoAtomIdxs()
    self.doublebonds_stereo = self.findDoubleBondsStereo()
    # calculate SPS
    self.score = self.calculateSpacialScore()
    # return nSPS
    if normalize:
      self.score /= self.mol.GetNumHeavyAtoms()

  def findStereoAtomIdxs(self, includeUnassigned=True):
    """Finds indices of atoms that are (pseudo)stereo/chiralcentres, in respect to the attached groups (does not account for double bond isomers)"""
    stereo_centers = Chem.FindMolChiralCenters(
      self.mol,
      includeUnassigned=includeUnassigned,
      includeCIP=False,
      useLegacyImplementation=False,
    )
    stereo_idxs = [atom_idx for atom_idx, _ in stereo_centers]
    return stereo_idxs

  def findDoubleBondsStereo(self):
    """Finds indeces of stereo double bond atoms (E/Z)"""
    db_stereo = {}
    for bond in self.mol.GetBonds():
      if bond.GetBondType() == Chem.rdchem.BondType.DOUBLE:
        db_stereo[(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())] = bond.GetStereo()
    return db_stereo

  def calculateSpacialScore(self):
    """Calculates the total spacial score for a molecule"""
    score = 0
    for atom in self.mol.GetAtoms():
      atom_idx = atom.GetIdx()
      self.hyb_score[atom_idx] = self._accountForHybridisation(atom)
      self.stereo_score[atom_idx] = self._accountForStereo(atom_idx)
      self.ring_score[atom_idx] = self._accountForRing(atom)
      self.bond_score[atom_idx] = self._accountForNeighbors(atom)
      score += self._calculateScoreForAtom(atom_idx)  # absolute score
    return score

  def _calculateScoreForAtom(self, atom_idx):
    """Calculates the total score for a single atom in a molecule"""
    atom_score = (self.hyb_score[atom_idx] * self.stereo_score[atom_idx] *
                  self.ring_score[atom_idx] * self.bond_score[atom_idx])
    return atom_score

  _hybridisations = defaultdict(lambda: 4)
  _hybridisations.update({
    Chem.HybridizationType.SP: 1,
    Chem.HybridizationType.SP2: 2,
    Chem.HybridizationType.SP3: 3
  })

  def _accountForHybridisation(self, atom):
    """Calculates the hybridisation score for a single atom in a molecule"""
    return self._hybridisations[atom.GetHybridization()]

  def _accountForStereo(self, atom_idx):
    """Calculates the stereo score for a single atom in a molecule"""
    if atom_idx in self.chiral_idxs:
      return 2
    for bond_atom_idxs, stereo in self.doublebonds_stereo.items():
      if stereo != Chem.BondStereo.STEREONONE and atom_idx in bond_atom_idxs:
        return 2
    return 1

  def _accountForRing(self, atom):
    """Calculates the ring score for a single atom in a molecule"""
    if atom.GetIsAromatic():  # aromatic rings are not promoted
      return 1
    if atom.IsInRing():
      return 2
    return 1

  def _accountForNeighbors(self, atom):
    """Calculates the neighbour score for a single atom in a molecule
        The second power allows to account for branching in the molecular structure"""
    return atom.GetDegree()**2



#  Copyright (C) 2000-2006  greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
""" a set of functions for interacting with databases

 When possible, it's probably preferable to use a _DbConnection.DbConnect_ object

"""

import sys
from io import StringIO

from rdkit.Dbase import DbInfo, DbModule
from rdkit.Dbase.DbResultSet import DbResultSet, RandomAccessDbResultSet


def _take(fromL, what):
  """ Given a list fromL, returns an iterator of the elements specified using their
    indices in the list what """
  return [fromL[x] for x in what]


def GetColumns(dBase, table, fieldString, user='sysdba', password='masterkey', join='', cn=None):
  """ gets a set of data from a table

      **Arguments**

       - dBase: database name

       - table: table name

       - fieldString: a string with the names of the fields to be extracted,
          this should be a comma delimited list

       - user and  password:

       - join: a join clause (omit the verb 'join')


      **Returns**

       - a list of the data

    """
  if not cn:
    cn = DbModule.connect(dBase, user, password)
  c = cn.cursor()
  cmd = 'select %s from %s' % (fieldString, table)
  if join:
    if join.strip().find('join') != 0:
      join = 'join %s' % (join)
    cmd += ' ' + join
  c.execute(cmd)
  return c.fetchall()


def GetData(dBase, table, fieldString='*', whereString='', user='sysdba', password='masterkey',
            removeDups=-1, join='', forceList=0, transform=None, randomAccess=1, extras=None,
            cn=None):
  """ a more flexible method to get a set of data from a table

      **Arguments**

       - fields: a string with the names of the fields to be extracted,
            this should be a comma delimited list

       - where: the SQL where clause to be used with the DB query

       - removeDups indicates the column which should be used to screen
          out duplicates.  Only the first appearance of a duplicate will
          be left in the dataset.

      **Returns**

        - a list of the data


      **Notes**

        - EFF: this isn't particularly efficient

    """
  if forceList and (transform is not None):
    raise ValueError('forceList and transform arguments are not compatible')
  if forceList and (not randomAccess):
    raise ValueError('when forceList is set, randomAccess must also be used')
  if removeDups > -1:
    forceList = True

  if not cn:
    cn = DbModule.connect(dBase, user, password)
  c = cn.cursor()
  cmd = 'select %s from %s' % (fieldString, table)
  if join:
    if join.strip().find('join') != 0:
      join = 'join %s' % (join)
    cmd += ' ' + join
  if whereString:
    if whereString.strip().find('where') != 0:
      whereString = 'where %s' % (whereString)
    cmd += ' ' + whereString

  if forceList:
    try:
      if not extras:
        c.execute(cmd)
      else:
        c.execute(cmd, extras)
    except Exception:
      sys.stderr.write('the command "%s" generated errors:\n' % (cmd))
      import traceback
      traceback.print_exc()
      return None
    if transform is not None:
      raise ValueError('forceList and transform arguments are not compatible')
    if not randomAccess:
      raise ValueError('when forceList is set, randomAccess must also be used')
    data = c.fetchall()
    if removeDups >= 0:
      seen = set()
      for entry in data[:]:
        if entry[removeDups] in seen:
          data.remove(entry)
        else:
          seen.add(entry[removeDups])
  else:
    if randomAccess:
      klass = RandomAccessDbResultSet
    else:
      klass = DbResultSet

    data = klass(c, cn, cmd, removeDups=removeDups, transform=transform, extras=extras)

  return data


def DatabaseToText(dBase, table, fields='*', join='', where='', user='sysdba', password='masterkey',
                   delim=',', cn=None):
  """ Pulls the contents of a database and makes a deliminted text file from them

      **Arguments**
        - dBase: the name of the DB file to be used

        - table: the name of the table to query

        - fields: the fields to select with the SQL query

        - join: the join clause of the SQL query
          (e.g. 'join foo on foo.bar=base.bar')

        - where: the where clause of the SQL query
          (e.g. 'where foo = 2' or 'where bar > 17.6')

        - user: the username for DB access

        - password: the password to be used for DB access

      **Returns**

        - the CSV data (as text)

    """
  if len(where) and where.strip().find('where') == -1:
    where = 'where %s' % (where)
  if len(join) and join.strip().find('join') == -1:
    join = 'join %s' % (join)
  sqlCommand = 'select %s from %s %s %s' % (fields, table, join, where)
  if not cn:
    cn = DbModule.connect(dBase, user, password)
  c = cn.cursor()
  c.execute(sqlCommand)
  headers = []
  colsToTake = []
  # the description field of the cursor carries around info about the columns
  #  of the table
  for i in range(len(c.description)):
    item = c.description[i]
    if item[1] not in DbInfo.sqlBinTypes:
      colsToTake.append(i)
      headers.append(item[0])

  lines = []
  lines.append(delim.join(headers))

  # grab the data
  results = c.fetchall()
  for res in results:
    d = _take(res, colsToTake)
    lines.append(delim.join([str(x) for x in d]))

  return '\n'.join(lines)


def TypeFinder(data, nRows, nCols, nullMarker=None):
  """

      finds the types of the columns in _data_

      if nullMarker is not None, elements of the data table which are
        equal to nullMarker will not count towards setting the type of
        their columns.

    """
  priorities = {float: 3, int: 2, str: 1, -1: -1}
  res = [None] * nCols
  for col in range(nCols):
    typeHere = [-1, 1]
    for row in range(nRows):
      d = data[row][col]
      if d is None:
        continue
      locType = type(d)
      if locType != float and locType != int:
        locType = str
        try:
          d = str(d)
        except UnicodeError as msg:
          print('cannot convert text from row %d col %d to a string' % (row + 2, col))
          print('\t>%s' % (repr(d)))
          raise UnicodeError(msg)
      else:
        typeHere[1] = max(typeHere[1], len(str(d)))
      if isinstance(d, str):
        if nullMarker is None or d != nullMarker:
          l = max(len(d), typeHere[1])
          typeHere = [str, l]
      else:
        try:
          fD = float(int(d))
        except OverflowError:
          locType = float
        else:
          if fD == d:
            locType = int
        if not isinstance(typeHere[0], str) and priorities[locType] > priorities[typeHere[0]]:
          typeHere[0] = locType
    res[col] = typeHere
  return res


def _AdjustColHeadings(colHeadings, maxColLabelLen):
  """ *For Internal Use*

      removes illegal characters from column headings
      and truncates those which are too long.

    """
  for i in range(len(colHeadings)):
    # replace unallowed characters and strip extra white space
    colHeadings[i] = colHeadings[i].strip()
    colHeadings[i] = colHeadings[i].replace(' ', '_')
    colHeadings[i] = colHeadings[i].replace('-', '_')
    colHeadings[i] = colHeadings[i].replace('.', '_')

    if len(colHeadings[i]) > maxColLabelLen:
      # interbase (at least) has a limit on the maximum length of a column name
      newHead = colHeadings[i].replace('_', '')
      newHead = newHead[:maxColLabelLen]
      print('\tHeading %s too long, changed to %s' % (colHeadings[i], newHead))
      colHeadings[i] = newHead
  return colHeadings


def GetTypeStrings(colHeadings, colTypes, keyCol=None):
  """  returns a list of SQL type strings
    """
  typeStrs = []
  for i in range(len(colTypes)):
    typ = colTypes[i]
    if typ[0] == float:
      typeStrs.append('%s double precision' % colHeadings[i])
    elif typ[0] == int:
      typeStrs.append('%s integer' % colHeadings[i])
    else:
      typeStrs.append('%s varchar(%d)' % (colHeadings[i], typ[1]))
    if colHeadings[i] == keyCol:
      typeStrs[-1] = '%s not null primary key' % (typeStrs[-1])
  return typeStrs


def _insertBlock(conn, sqlStr, block, silent=False):
  try:
    conn.cursor().executemany(sqlStr, block)
  except Exception:
    res = 0
    conn.commit()
    for row in block:
      try:
        conn.cursor().execute(sqlStr, tuple(row))
        res += 1
      except Exception:
        if not silent:
          import traceback
          traceback.print_exc()
          print('insert failed:', sqlStr)
          print('\t', repr(row))
      else:
        conn.commit()
  else:
    res = len(block)
  return res


def _AddDataToDb(dBase, table, user, password, colDefs, colTypes, data, nullMarker=None,
                 blockSize=100, cn=None):
  """ *For Internal Use*

      (drops and) creates a table and then inserts the values

    """
  if not cn:
    cn = DbModule.connect(dBase, user, password)
  c = cn.cursor()
  try:
    c.execute('drop table %s' % (table))
  except Exception:
    print('cannot drop table %s' % (table))
  try:
    sqlStr = 'create table %s (%s)' % (table, colDefs)
    c.execute(sqlStr)
  except Exception:
    print('create table failed: ', sqlStr)
    print('here is the exception:')
    import traceback
    traceback.print_exc()
    return
  cn.commit()
  c = None

  block = []
  entryTxt = [DbModule.placeHolder] * len(data[0])
  dStr = ','.join(entryTxt)
  sqlStr = 'insert into %s values (%s)' % (table, dStr)
  nDone = 0
  for row in data:
    entries = [None] * len(row)
    for col in range(len(row)):
      if row[col] is not None and \
         (nullMarker is None or row[col] != nullMarker):
        if colTypes[col][0] == float:
          entries[col] = float(row[col])
        elif colTypes[col][0] == int:
          entries[col] = int(row[col])
        else:
          entries[col] = str(row[col])
      else:
        entries[col] = None
    block.append(tuple(entries))
    if len(block) >= blockSize:
      nDone += _insertBlock(cn, sqlStr, block)
      # note: in Python 3.12 `cn.autocommit != True`
      # is different from `not cn.autocommit` (GH #7009)
      if not hasattr(cn, 'autocommit') or cn.autocommit != True:
        cn.commit()
      block = []
  if len(block):
    nDone += _insertBlock(cn, sqlStr, block)
  # note: in Python 3.12 `cn.autocommit != True`
  # is different from `not cn.autocommit` (GH #7009)
  if not hasattr(cn, 'autocommit') or cn.autocommit != True:
    cn.commit()


def TextFileToDatabase(dBase, table, inF, delim=',', user='sysdba', password='masterkey',
                       maxColLabelLen=31, keyCol=None, nullMarker=None):
  """loads the contents of the text file into a database.

      **Arguments**

        - dBase: the name of the DB to use

        - table: the name of the table to create/overwrite

        - inF: the file like object from which the data should
          be pulled (must support readline())

        - delim: the delimiter used to separate fields

        - user: the user name to use in connecting to the DB

        - password: the password to use in connecting to the DB

        - maxColLabelLen: the maximum length a column label should be
          allowed to have (truncation otherwise)

        - keyCol: the column to be used as an index for the db

      **Notes**

        - if _table_ already exists, it is destroyed before we write
          the new data

        - we assume that the first row of the file contains the column names

    """
  table.replace('-', '_')
  table.replace(' ', '_')

  colHeadings = inF.readline().split(delim)
  _AdjustColHeadings(colHeadings, maxColLabelLen)
  nCols = len(colHeadings)
  data = []
  inL = inF.readline()
  while inL:
    inL = inL.replace('\r', '')
    inL = inL.replace('\n', '')
    splitL = inL.split(delim)
    if len(splitL) != nCols:
      print('>>>', repr(inL))
      assert len(splitL) == nCols, 'unequal length'
    tmpVect = []
    for entry in splitL:
      try:
        val = int(entry)
      except ValueError:
        try:
          val = float(entry)
        except ValueError:
          val = entry
      tmpVect.append(val)
    data.append(tmpVect)
    inL = inF.readline()
  nRows = len(data)

  # determine the types of each column
  colTypes = TypeFinder(data, nRows, nCols, nullMarker=nullMarker)
  typeStrs = GetTypeStrings(colHeadings, colTypes, keyCol=keyCol)
  colDefs = ','.join(typeStrs)

  _AddDataToDb(dBase, table, user, password, colDefs, colTypes, data, nullMarker=nullMarker)


def DatabaseToDatabase(fromDb, fromTbl, toDb, toTbl, fields='*', join='', where='', user='sysdba',
                       password='masterkey', keyCol=None, nullMarker='None'):
  """

     FIX: at the moment this is a hack

    """
  sio = StringIO()
  sio.write(
    DatabaseToText(fromDb, fromTbl, fields=fields, join=join, where=where, user=user,
                   password=password))
  sio.seek(0)
  TextFileToDatabase(toDb, toTbl, sio, user=user, password=password, keyCol=keyCol,
                     nullMarker=nullMarker)


if __name__ == '__main__':  # pragma: nocover
  sio = StringIO()
  sio.write('foo,bar,baz\n')
  sio.write('1,2,3\n')
  sio.write('1.1,4,5\n')
  sio.write('4,foo,6\n')
  sio.seek(0)
  import os

  from rdkit import RDConfig
  dirLoc = os.path.join(RDConfig.RDCodeDir, 'Dbase', 'TEST.GDB')

  TextFileToDatabase(dirLoc, 'fromtext', sio)



'''

Script will test the RDkit python code for conformance with the agreed format using
yapf.

For each Python file that is found in $RDBASE (excluding the build and External
directories), yapf is used with the style configuration in $RDBASE/setup.cfg.
If a change is required, the difference is printed. At the end of the process,
all non-conformant files are listed and the required yapf command(s) printed.

If changes are found, the script will exit with error code 1, otherwise 0.

'''

import os
import sys

from yapf.yapflib.yapf_api import FormatCode

rdbase = os.environ.get('RDBASE', '')
styleConfig = os.path.join(rdbase, 'setup.cfg')

excludeDirs = [
  os.path.join(rdbase, 'build'),
  os.path.join(rdbase, 'External'),
]


def pythonFiles(dirname=rdbase):
  """ Find all python files below directory dirname """
  for root, _, files in os.walk(dirname):
    if any(root.startswith(d) for d in excludeDirs):
      continue
    for file in files:
      if file.endswith(".py"):
        yield os.path.join(root, file)


def yapfChanges(filename):
  """ Use yapf with the default settings to format file filename """
  try:
    with open(filename) as f:
      codeBefore = f.read()
  except UnicodeError:
    with open(filename, encoding='latin-1') as f:
      codeBefore = f.read()
  try:
    changes, changed = FormatCode(codeBefore, style_config=styleConfig, print_diff=True,
                                  filename=filename)
  except Exception:
    print(filename)
    raise
  if changed:
    print(changes)
  return changed


if __name__ == "__main__":
  changedFiles = []
  for s in pythonFiles():
    if yapfChanges(s):
      changedFiles.append(s)
  print()
  if changedFiles:
    print('yapf will make changes to the following files:')
    print('\n'.join(sorted(changedFiles)))
    print('To apply the required changes to your code use the following command(s)')
    for s in sorted(set(s.replace(rdbase, '').split(os.sep)[1] for s in changedFiles)):
      print('yapf --style $RDBASE/setup.cfg --in-place --recursive $RDBASE/{0}'.format(s))
    sys.exit(1)
  print('Code complies with the agreed formatting rules.')
  sys.exit(0)



# Copyright (C) 2022 Schrödinger, LLC
# This file is part of the RDKit.
# The contents are covered by the terms of the BSD license
# which is included in the file license.txt, found at the root
# of the RDKit source tree.
"""
Generate a unique hash code for a molecule based on chemistry. If two
molecules are chemically "the same", they should have the same hash.

Using molhash adds value beyond using SMILES because it:

* Ignores SMILES features that are not chemically meaningful
(e.g. atom map numbers)
* Canonicalizes enhanced stereochemistry groups. For example
`C[C@H](O)CC |&1:1|` and `C[C@@H](O)CC |&1:1|` have the same
molhash
* Canonicalizes S group data (for example, polymer data)

There are two hash schemes, the default, and one in which
tautomers are considered equivalent.

"""

import enum
import hashlib
import json
import logging
import re
from typing import Dict, Iterable, Optional

from rdkit import Chem
from rdkit.Chem import rdMolHash

ATOM_PROP_MAP_NUMBER = 'molAtomMapNumber'

logger = logging.getLogger(__name__)

DEFAULT_CXFLAG = (Chem.CXSmilesFields.CX_ATOM_LABELS | Chem.CXSmilesFields.CX_ENHANCEDSTEREO
                  | Chem.CXSmilesFields.CX_BOND_ATROPISOMER)

ENHANCED_STEREO_GROUP_REGEX = re.compile(r'((?:a|[&o]\d+):\d+(?:,\d+)*)')

ENHANCED_STEREO_GROUP_WEIGHTS = {
  Chem.StereoGroupType.STEREO_AND: 1000,
  Chem.StereoGroupType.STEREO_OR: 2000,
  Chem.StereoGroupType.STEREO_ABSOLUTE: 3000,
}

EMPTY_MOL_TAUTOMER_HASH = "_0_0"


class HashLayer(enum.Enum):
  """
    :cvar CANONICAL_SMILES: RDKit canonical SMILES (excluding enhanced stereo)
    :cvar ESCAPE: arbitrary other information to be incorporated
    :cvar FORMULA: a simple molecular formula for the molecule
    :cvar NO_STEREO_SMILES: RDKit canonical SMILES with all stereo removed
    :cvar SGROUP_DATA: canonicalization of all SGroups data present
    :cvar TAUTOMER_HASH: SMILES-like representation for a generic tautomer form
    :cvar NO_STEREO_TAUTOMER_HASH: the above tautomer hash lacking all stereo
    """
  CANONICAL_SMILES = enum.auto()
  ESCAPE = enum.auto()
  FORMULA = enum.auto()
  NO_STEREO_SMILES = enum.auto()
  NO_STEREO_TAUTOMER_HASH = enum.auto()
  SGROUP_DATA = enum.auto()
  TAUTOMER_HASH = enum.auto()


@enum.unique
class HashScheme(enum.Enum):
  """
    Which hash layers to use to when deduplicating molecules

    Typically the "ALL_LAYERS" scheme is used, but some users may want
    the "TAUTOMER_INSENSITIVE_LAYERS" scheme.

    :cvar ALL_LAYERS: most strict hash scheme utilizing all layers
    :cvar STEREO_INSENSITIVE_LAYERS: excludes stereo sensitive layers
    :cvar TAUTOMER_INSENSITIVE_LAYERS: excludes tautomer sensitive layers
    """
  ALL_LAYERS = tuple(HashLayer)
  STEREO_INSENSITIVE_LAYERS = (
    HashLayer.ESCAPE,
    HashLayer.FORMULA,
    HashLayer.NO_STEREO_SMILES,
    HashLayer.NO_STEREO_TAUTOMER_HASH,
    HashLayer.SGROUP_DATA,
  )
  TAUTOMER_INSENSITIVE_LAYERS = (
    HashLayer.ESCAPE,
    HashLayer.FORMULA,
    HashLayer.NO_STEREO_TAUTOMER_HASH,
    HashLayer.SGROUP_DATA,
    HashLayer.TAUTOMER_HASH,
  )


def GetMolHash(all_layers, hash_scheme: HashScheme = HashScheme.ALL_LAYERS) -> str:
  """
    Generate a molecular hash using a specified set of layers.

    :param all_layers: a dictionary of layers
    :param hash_scheme: enum encoding information layers for the hash
    :return: hash for the given scheme constructed from the input layers
    """

  h = hashlib.sha1()
  for layer in hash_scheme.value:
    if layer in all_layers:
      h.update(all_layers[layer].encode())
  return h.hexdigest()


def GetMolLayers(original_molecule: Chem.rdchem.Mol, data_field_names: Optional[Iterable] = None,
                 escape: Optional[str] = None, cxflag=DEFAULT_CXFLAG,
                 enable_tautomer_hash_v2=False) -> Dict[HashLayer, str]:
  """
    Generate layers of data about that could be used to identify a molecule

    :param original_molecule: molecule to obtain canonicalization layers from
    :param data_field_names: optional sequence of names of SGroup DAT fields which
       will be included in the hash.
    :param escape: optional field which can contain arbitrary information
    :param enable_tautomer_hash_v2: use v2 of the tautomer hash
    :return: dictionary of HashLayer enum to calculated hash
    """
  # Work on a copy with all non-stereogenic Hydrogens removed
  mol = _RemoveUnnecessaryHs(original_molecule, preserve_stereogenic_hs=True)
  _StripAtomMapLabels(mol)

  # we only canonicalize the stereo groups if we are actually outputting enhanced stereo:
  if cxflag & Chem.CXSmilesFields.CX_ENHANCEDSTEREO:
    Chem.CanonicalizeEnhancedStereo(mol)

  formula = rdMolHash.MolHash(mol, rdMolHash.HashFunction.MolFormula)

  ps = Chem.SmilesWriteParams()
  cxsmiles = Chem.MolToCXSmiles(mol, ps, cxflag)

  tautomer_hash = GetStereoTautomerHash(mol, cxflag=cxflag,
                                        enable_tautomer_hash_v2=enable_tautomer_hash_v2)

  sgroup_data = _CanonicalizeSGroups(mol, dataFieldNames=data_field_names)

  no_stereo_tautomer_hash, no_stereo_smiles = GetNoStereoLayers(
    mol, enable_tautomer_hash_v2=enable_tautomer_hash_v2)

  return {
    HashLayer.CANONICAL_SMILES: cxsmiles,
    HashLayer.ESCAPE: escape or "",
    HashLayer.FORMULA: formula,
    HashLayer.NO_STEREO_SMILES: no_stereo_smiles,
    HashLayer.NO_STEREO_TAUTOMER_HASH: no_stereo_tautomer_hash,
    HashLayer.SGROUP_DATA: sgroup_data,
    HashLayer.TAUTOMER_HASH: tautomer_hash,
  }


def _StripAtomMapLabels(mol):
  for at in mol.GetAtoms():
    at.ClearProp(ATOM_PROP_MAP_NUMBER)


def _RemoveUnnecessaryHs(rdk_mol, preserve_stereogenic_hs=False):
  """
    removes hydrogens that are not necessary for the registration hash, and
    preserves hydrogen isotopes
    """
  remove_hs_params = Chem.RemoveHsParameters()
  remove_hs_params.updateExplicitCount = True
  remove_hs_params.removeDefiningBondStereo = not preserve_stereogenic_hs
  edited_mol = Chem.rdmolops.RemoveHs(rdk_mol, remove_hs_params, sanitize=False)
  edited_mol.UpdatePropertyCache(False)
  return edited_mol


def GetStereoTautomerHash(molecule, cxflag=DEFAULT_CXFLAG, enable_tautomer_hash_v2=False):
  if molecule.GetNumAtoms() == 0:
    return EMPTY_MOL_TAUTOMER_HASH

  # SHARED-7909: workaround for https://github.com/rdkit/rdkit/issues/4234
  # This can be removed once we update to an RDKit version without this bug.
  no_h_mol = _RemoveUnnecessaryHs(molecule)
  no_h_mol.UpdatePropertyCache(False)

  # setting useCxSmiles param value to always include enhanced stereo info
  useCxSmiles = True
  cx_flags_to_skip = Chem.CXSmilesFields.CX_ALL ^ cxflag
  if enable_tautomer_hash_v2:
    hash_func = rdMolHash.HashFunction.HetAtomTautomerv2
  else:
    hash_func = rdMolHash.HashFunction.HetAtomTautomer

  hash_with_cxExtensions = rdMolHash.MolHash(no_h_mol, hash_func, useCxSmiles, cx_flags_to_skip)

  return hash_with_cxExtensions


def GetNoStereoLayers(mol, enable_tautomer_hash_v2=False):
  # SHARED-7909: Strip all Hs, including stereogenic ones.
  no_stereo_mol = _RemoveUnnecessaryHs(mol)
  no_stereo_mol.UpdatePropertyCache(False)
  Chem.rdmolops.RemoveStereochemistry(no_stereo_mol)

  if enable_tautomer_hash_v2:
    hash_func = rdMolHash.HashFunction.HetAtomTautomerv2
  else:
    hash_func = rdMolHash.HashFunction.HetAtomTautomer
  no_stereo_tautomer_hash = rdMolHash.MolHash(no_stereo_mol, hash_func)

  no_stereo_smiles = rdMolHash.MolHash(no_stereo_mol, rdMolHash.HashFunction.CanonicalSmiles)
  return (no_stereo_tautomer_hash, no_stereo_smiles)


def _GetCanonicalAtomRanksAndBonds(mol, useSmilesOrdering=True):
  """
    returns a 2-tuple with:

    1. the canonical ranks of a molecule's atoms
    2. the bonds expressed as (canonical_atom_rank_1,canonical_atom_rank_2) where
       canonical_atom_rank_1 < canonical_atom_rank_2

    If useSmilesOrdering is True then the atom indices here correspond to the order of
    the atoms in the canonical SMILES, otherwise just the canonical atom order is used.
    useSmilesOrdering=True is a bit slower, but it allows the output to be linked to the
    canonical SMILES, which can be useful.

    """
  if mol.GetNumAtoms() == 0:
    return [], []
  if not useSmilesOrdering:
    atRanks = list(Chem.CanonicalRankAtoms(mol))
  else:
    smi = Chem.MolToSmiles(mol)
    ordertxt = mol.GetProp("_smilesAtomOutputOrder")
    smiOrder = [int(x) for x in ordertxt[1:-1].split(",") if x]
    atRanks = [0] * len(smiOrder)
    for i, idx in enumerate(smiOrder):
      atRanks[idx] = i
  # get the bonds in canonical form
  bndOrder = []
  for bnd in mol.GetBonds():
    bo = atRanks[bnd.GetBeginAtomIdx()]
    eo = atRanks[bnd.GetEndAtomIdx()]
    if bo > eo:
      bo, eo = eo, bo
    bndOrder.append((bo, eo))
  return atRanks, bndOrder


def _CanonicalizeDataSGroup(sg, atRanks, bndOrder, fieldNames=("Atrop", ),
                            sortAtomAndBondOrder=True):
  """
    NOTES: if sortAtomAndBondOrder is true then the atom and bond lists will
    be sorted. This assumes that the order of the atoms in that list is not
    important

    """
  if sg.GetProp("TYPE") != "DAT" or not sg.HasProp("FIELDNAME"):
    return None

  fieldName = sg.GetProp("FIELDNAME")
  if fieldName not in fieldNames:
    return None

  data = sg.GetStringVectProp("DATAFIELDS")
  if len(data) > 1:
    raise ValueError("cannot canonicalize data groups with multiple data fields")
  data = data[0]
  ats = tuple(atRanks[x] for x in sg.GetAtoms())
  if sortAtomAndBondOrder:
    ats = tuple(sorted(ats))
  bnds = tuple(bndOrder[x] for x in sg.GetBonds())
  if sortAtomAndBondOrder:
    bnds = tuple(sorted(bnds))

  res = dict(fieldName=fieldName, atom=ats, bonds=bnds, value=data)
  return res


def _GetCanononicalBondRep(bond, atomRanks):
  aid1 = bond.GetBeginAtomIdx()
  aid2 = bond.GetEndAtomIdx()
  if atomRanks[aid1] > atomRanks[aid2] or (atomRanks[aid1] == atomRanks[aid2] and aid1 > aid2):
    aid1, aid2 = aid2, aid1
  return (aid1, aid2)


def _CanonicalizeSRUSGroup(mol, sg, atRanks, bndOrder, sortAtomAndBondOrder):
  """
    NOTES: if sortAtomAndBondOrder is true then the atom and bond lists will be sorted.
    This assumes that the ordering of those lists is not important

    """
  if sg.GetProp("TYPE") != "SRU":
    return None
  ats = tuple(atRanks[x] for x in sg.GetAtoms())
  bnds = tuple(bndOrder[x] for x in sg.GetBonds())
  if sortAtomAndBondOrder:
    ats = tuple(sorted(ats))
    bnds = tuple(sorted(bnds))
  props = sg.GetPropsAsDict()
  res = dict(
    type="SRU",
    atoms=ats,
    bonds=bnds,
    index=props.get("index", 0),
    connect=props.get("CONNECT", ""),
    label=props.get("LABEL", ""),
  )
  if "PARENT" in props:
    res["PARENT"] = props["PARENT"]
  if "XBHEAD" in props:
    xbhbonds = tuple(
      _GetCanononicalBondRep(mol.GetBondWithIdx(x), atRanks) for x in props["XBHEAD"])
    if sortAtomAndBondOrder:
      xbhbonds = tuple(sorted(xbhbonds))
    res["XBHEAD"] = xbhbonds
  if "XBCORR" in props:
    xbcorrbonds = tuple(
      _GetCanononicalBondRep(mol.GetBondWithIdx(x), atRanks) for x in props["XBCORR"])
    if len(xbcorrbonds) % 2:
      raise ValueError("XBCORR should have 2N bonds")
    if sortAtomAndBondOrder:
      # these are pairs, so we need to sort them as such
      tmp = []
      for i in range(0, len(xbcorrbonds), 2):
        b1, b2 = xbcorrbonds[i], xbcorrbonds[i + 1]
        if b1 > b2:
          b1, b2 = b2, b1
        tmp.append((b1, b2))
      xbcorrbonds = tuple(sorted(tmp))
    res["XBCORR"] = xbcorrbonds
  return res


def _CanonicalizeCOPSGroup(sg, atRanks, sortAtomAndBondOrder):
  """
    NOTES: if sortAtomAndBondOrder is true then the atom and bond lists will be sorted.
    This assumes that the ordering of those lists is not important
    """
  if sg.GetProp("TYPE") != "COP":
    return None
  ats = tuple(atRanks[x] for x in sg.GetAtoms())
  if sortAtomAndBondOrder:
    ats = tuple(sorted(ats))
  props = sg.GetPropsAsDict()
  # we are intentionally not doing the label here. Marvin adds one, biovia draw does not
  res = dict(type="COP", atoms=ats, index=props.get("index", 0))
  return res


def _CanonicalizeSGroups(mol, dataFieldNames=None, sortAtomAndBondOrder=True):
  """
    NOTES: if sortAtomAndBondOrder is true then the atom and bond lists will be sorted.
    This assumes that the ordering of those lists is not important
    """
  dataFieldNames = dataFieldNames or ["Atrop"]
  atRanks, bndOrder = _GetCanonicalAtomRanksAndBonds(mol)
  res = []
  for sg in Chem.GetMolSubstanceGroups(mol):
    lres = None
    if sg.GetProp("TYPE") == "DAT":
      lres = _CanonicalizeDataSGroup(sg, atRanks, bndOrder, dataFieldNames, sortAtomAndBondOrder)
    elif sg.GetProp("TYPE") == "SRU":
      lres = _CanonicalizeSRUSGroup(mol, sg, atRanks, bndOrder, sortAtomAndBondOrder)
    elif sg.GetProp("TYPE") == "COP":
      lres = _CanonicalizeCOPSGroup(sg, atRanks, sortAtomAndBondOrder)

    if lres is not None:
      res.append(lres)
  if len(res) > 1:
    # we need to sort them, but dicts don't let you do that directly:
    tres = sorted(tuple(x.items()) for x in res)
    res = tuple(dict(x) for x in tres)
    # update "index" and "PARENT" props
    idxmap = {}
    for i, itm in enumerate(res):
      if "index" in itm:
        idxmap[itm["index"]] = i + 1
        itm["index"] = i + 1
    for i, itm in enumerate(res):
      if "PARENT" in itm:
        itm["PARENT"] = idxmap[itm["PARENT"]]

  return json.dumps(res)


class EnhancedStereoUpdateMode(enum.Enum):
  ADD_WEIGHTS = enum.auto()
  REMOVE_WEIGHTS = enum.auto()
